# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for power controls."""

from django.forms import Form
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse

from plinth import app as app_module
from plinth import package
from plinth.views import AppView

from . import privileged


class PowerAppView(AppView):
    """Show power app main page."""

    app_id = 'power'
    template_name = 'power.html'

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        is_busy = package.is_package_manager_busy()
        context['pkg_manager_is_busy'] = is_busy
        return context


def restart(request):
    """Serve start confirmation page."""
    form = None

    if request.method == 'POST':
        privileged.restart(_run_in_background=True)
        return redirect(reverse('apps'))

    app = app_module.App.get('power')
    form = Form(prefix='power')
    is_busy = package.is_package_manager_busy()
    return TemplateResponse(
        request, 'power_restart.html', {
            'title': app.info.name,
            'form': form,
            'manual_page': app.info.manual_page,
            'pkg_manager_is_busy': is_busy
        })


def shutdown(request):
    """Serve shutdown confirmation page."""
    form = None

    if request.method == 'POST':
        privileged.shutdown(_run_in_background=True)
        return redirect(reverse('apps'))

    app = app_module.App.get('power')
    form = Form(prefix='power')
    is_busy = package.is_package_manager_busy()
    return TemplateResponse(
        request, 'power_shutdown.html', {
            'title': app.info.name,
            'form': form,
            'manual_page': app.info.manual_page,
            'pkg_manager_is_busy': is_busy
        })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Single Sign On module.
"""

from django.urls import re_path
from stronghold.decorators import public

from plinth.utils import non_admin_view

from .views import CaptchaView, SSOLoginView, refresh

urlpatterns = [
    re_path(r'^accounts/sso/login/$', public(SSOLoginView.as_view()),
            name='sso-login'),
    re_path(r'^accounts/sso/refresh/$', non_admin_view(refresh),
            name='sso-refresh'),

    # Locked URL from django-axes
    re_path(r'accounts/sso/login/locked/$', public(CaptchaView.as_view()),
            name='locked_out'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Syncthing."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.users import add_user_to_share_group
from plinth.modules.users import privileged as users_privileged
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Syncthing is an application to synchronize files across multiple '
      'devices, e.g. your desktop computer and mobile phone.  Creation, '
      'modification, or deletion of files on one device will be automatically '
      'replicated on all other devices that also run Syncthing.'),
    format_lazy(
        _('Running Syncthing on {box_name} provides an extra synchronization '
          'point for your data that is available most of the time, allowing '
          'your devices to synchronize more often.  {box_name} runs a single '
          'instance of Syncthing that may be used by multiple users.  Each '
          'user\'s set of devices may be synchronized with a distinct set of '
          'folders.  The web interface on {box_name} is only available for '
          'users belonging to the "admin" or "syncthing-access" group.'),
        box_name=_(cfg.box_name)),
]

SYSTEM_USER = 'syncthing'


class SyncthingApp(app_module.App):
    """FreedomBox app for Syncthing."""

    app_id = 'syncthing'

    _version = 7

    DAEMON = 'syncthing@syncthing'

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        self.groups = {
            'syncthing-access': _('Administer Syncthing application')
        }

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Syncthing'), icon_filename='syncthing',
                               short_description=_('File Synchronization'),
                               description=_description,
                               manual_page='Syncthing',
                               clients=manifest.clients,
                               donation_url='https://syncthing.net/donations/')
        self.add(info)

        menu_item = menu.Menu('menu-syncthing', info.name,
                              info.short_description, info.icon_filename,
                              'syncthing:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-syncthing', info.name,
                                      short_description=info.short_description,
                                      icon=info.icon_filename,
                                      url='/syncthing/', clients=info.clients,
                                      login_required=True,
                                      allowed_groups=list(self.groups))
        self.add(shortcut)

        packages = Packages('packages-syncthing', ['syncthing'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-syncthing', [
            '/etc/apache2/conf-available/syncthing-plinth.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-syncthing-web', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall = Firewall('firewall-syncthing-ports', info.name,
                            ports=['syncthing'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-syncthing', ['8384'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-syncthing', 'syncthing-plinth',
                              urls=['https://{host}/syncthing/'])
        self.add(webserver)

        daemon = Daemon('daemon-syncthing', self.DAEMON)
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-syncthing',
                                          [SYSTEM_USER], self.groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-syncthing',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        add_user_to_share_group(SYSTEM_USER, SyncthingApp.DAEMON)

        if not old_version:
            self.enable()

        privileged.setup_config()

        if old_version == 1 and self.is_enabled():
            self.get_component('firewall-syncthing-ports').enable()

        if old_version and old_version <= 3:
            # rename LDAP and Django group
            old_groupname = 'syncthing'
            new_groupname = 'syncthing-access'

            users_privileged.rename_group(old_groupname, new_groupname)

            from django.contrib.auth.models import Group
            Group.objects.filter(name=old_groupname).update(name=new_groupname)

            # update web shares to have new group name
            from plinth.modules import sharing
            shares = sharing.list_shares()
            for share in shares:
                if old_groupname in share['groups']:
                    new_groups = share['groups']
                    new_groups.remove(old_groupname)
                    new_groups.append(new_groupname)

                    name = share['name']
                    sharing.remove_share(name)
                    sharing.add_share(name, share['path'], new_groups,
                                      share['is_public'])

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Tests for diagnostic check data type."""

import json

import pytest

from plinth.diagnostic_check import (CheckJSONDecoder, CheckJSONEncoder,
                                     DiagnosticCheck, Result)


def test_result():
    """Test result enum type."""
    assert Result.__members__['ERROR'].name == 'ERROR'
    assert Result.__members__['ERROR'].value == 'error'
    assert Result.NOT_DONE == 'not_done'
    assert Result.PASSED == 'passed'
    assert Result.WARNING == 'warning'
    assert Result.FAILED == 'failed'
    assert Result.ERROR == 'error'


def test_diagnostic_check():
    """Test the diagnostic check data class."""
    with pytest.raises(TypeError):
        DiagnosticCheck()  # pylint: disable=E1120

    check = DiagnosticCheck('some-check-id', 'sample check')
    assert check.check_id == 'some-check-id'
    assert check.description == 'sample check'
    assert check.translated_description == 'sample check'
    assert check.result == Result.NOT_DONE
    assert check.component_id is None
    assert not check.parameters

    check = DiagnosticCheck('some-check-id', 'sample check', Result.PASSED)
    assert check.result == Result.PASSED
    assert not check.parameters

    check = DiagnosticCheck('some-check-id', 'sample check', Result.PASSED,
                            {'key': 'value'})
    assert check.parameters['key'] == 'value'

    check = DiagnosticCheck('some-check-id', 'sample check', Result.FAILED, {},
                            'some-component')
    assert check.component_id == 'some-component'


def test_translate():
    """Test formatting the translated description."""
    check = DiagnosticCheck('some-check-id', 'sample check {key}',
                            Result.FAILED, {'key': 'value'})
    assert check.translated_description == 'sample check value'

    check = DiagnosticCheck('some-check-id', 'sample check {missing}',
                            Result.PASSED, {'key': 'value'})
    assert check.translated_description == 'sample check ?missing?'


def test_json_encoder_decoder():
    """Test encoding and decoding as JSON."""
    check = DiagnosticCheck('some-check-id', 'sample check', Result.PASSED)
    check_json = json.dumps(check, cls=CheckJSONEncoder)
    for string in [
            '"check_id": "some-check-id"', '"description": "sample check"',
            '"result": "passed"', '"parameters": {}', '"component_id": null',
            '"__class__": "DiagnosticCheck"'
    ]:
        assert string in check_json

    decoded_check = json.loads(check_json, cls=CheckJSONDecoder)
    assert decoded_check == check



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for clients module.
"""

from plinth import clients
from plinth.modules.deluge.manifest import clients as deluge_clients
from plinth.modules.infinoted.manifest import clients as infinoted_clients
from plinth.modules.quassel.manifest import clients as quassel_clients
from plinth.modules.syncthing.manifest import clients as syncthing_clients
from plinth.modules.tor.manifest import clients as tor_clients


def test_of_type_web():
    """Test filtering clients of type web."""
    assert clients.of_type(syncthing_clients, 'web')
    assert not clients.of_type(quassel_clients, 'web')


def test_of_type_mobile():
    """Test filtering clients of type mobile."""
    assert clients.of_type(syncthing_clients, 'mobile')
    assert not clients.of_type(infinoted_clients, 'mobile')


def test_of_type_desktop():
    """Test filtering clients of type desktop."""
    assert clients.of_type(syncthing_clients, 'desktop')
    assert not clients.of_type(deluge_clients, 'desktop')


def test_of_type_package():
    """Test filtering clients of type package."""
    assert clients.of_type(syncthing_clients, 'package')
    assert not clients.of_type(tor_clients, 'package')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for firewall app component.
"""

from unittest.mock import call, patch

import pytest

from plinth.app import App
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)


@pytest.fixture(name='empty_firewall_list', autouse=True)
def fixture_empty_firewall_list():
    """Remove all entries in firewall list before starting a test."""
    Firewall._all_firewall_components = {}


def test_init_without_arguments():
    """Test initializing the component without arguments."""
    with pytest.raises(ValueError):
        Firewall(None)

    firewall = Firewall('test-component')
    assert firewall.component_id == 'test-component'
    assert firewall.name is None
    assert firewall.ports == []
    assert not firewall.is_external
    assert len(Firewall._all_firewall_components.items()) == 1
    assert Firewall._all_firewall_components['test-component'] == firewall


def test_init():
    """Test initializing the component."""
    firewall = Firewall('test-component', 'test-name',
                        ['test-port1', 'test-port2'], is_external=True)
    assert firewall.name == 'test-name'
    assert firewall.ports == ['test-port1', 'test-port2']
    assert firewall.is_external


@patch('plinth.modules.firewall.get_port_details')
def test_port_details(get_port_details):
    """Test retrieving port details for a firewall component."""
    return_values = {
        'test-port1': [(1234, 'tcp')],
        'test-port2': [(5678, 'udp')]
    }

    def get_port_details_side_effect(port):
        return return_values[port]

    get_port_details.side_effect = get_port_details_side_effect
    firewall = Firewall('test-component', ports=['test-port1', 'test-port2'])
    assert firewall.ports_details == [{
        'name': 'test-port1',
        'details': [(1234, 'tcp')]
    }, {
        'name': 'test-port2',
        'details': [(5678, 'udp')]
    }]


@patch('plinth.modules.firewall.add_service')
@patch('plinth.modules.firewall.get_enabled_services')
def test_enable(get_enabled_services, add_service):
    """Test enabling a firewall component."""

    def get_enabled_services_side_effect(zone):
        return {'internal': ['test-port1'], 'external': ['test-port2']}[zone]

    get_enabled_services.side_effect = get_enabled_services_side_effect
    # Internal
    firewall = Firewall('test-firewall-1', ports=['test-port1', 'test-port2'],
                        is_external=False)
    firewall.enable()
    calls = [call('test-port2', zone='internal')]
    add_service.assert_has_calls(calls)

    # External
    add_service.reset_mock()
    firewall = Firewall('test-firewall-2', ports=['test-port1', 'test-port2'],
                        is_external=True)
    firewall.enable()
    calls = [
        call('test-port1', zone='external'),
        call('test-port2', zone='internal')
    ]
    add_service.assert_has_calls(calls)


@patch('plinth.modules.firewall.remove_service')
@patch('plinth.modules.firewall.add_service')
@patch('plinth.modules.firewall.get_enabled_services')
def test_disable(get_enabled_services, add_service, remove_service):
    """Test disabling a firewall component."""
    Firewall('firewall-1', ports=['test-port1'], is_external=False)
    Firewall('firewall-2', ports=['test-port2'], is_external=False).enable()
    Firewall('firewall-3', ports=['test-port4'], is_external=True)
    Firewall('firewall-4', ports=['test-port5'], is_external=True).enable()

    def get_enabled_services_side_effect(zone):
        return {
            'internal': ['test-port1', 'test-port2'],
            'external': ['test-port4', 'test-port5']
        }[zone]

    get_enabled_services.side_effect = get_enabled_services_side_effect
    all_ports = [
        'test-port1', 'test-port2', 'test-port3', 'test-port4', 'test-port5',
        'test-port6'
    ]
    # Internal
    firewall = Firewall('test-firewall-1', ports=all_ports, is_external=False)
    firewall.disable()
    calls = [call('test-port1', zone='internal')]
    remove_service.assert_has_calls(calls)

    # External
    remove_service.reset_mock()
    firewall = Firewall('test-firewall-2', ports=all_ports, is_external=True)
    firewall.disable()
    calls = [
        call('test-port1', zone='internal'),
        call('test-port4', zone='external')
    ]
    remove_service.assert_has_calls(calls)


@patch('plinth.modules.firewall.get_port_details')
@patch('plinth.modules.firewall.get_enabled_services')
def test_diagnose(get_enabled_services, get_port_details):
    """Test diagnosing open/closed firewall ports."""

    def get_port_details_side_effect(port):
        return {
            'test-port1': [(1234, 'tcp'), (1234, 'udp')],
            'test-port2': [(2345, 'udp')],
            'test-port3': [(3456, 'tcp')],
            'test-port4': [(4567, 'udp')]
        }[port]

    def get_enabled_services_side_effect(zone):
        return {
            'internal': ['test-port1', 'test-port3'],
            'external': ['test-port2', 'test-port3']
        }[zone]

    get_enabled_services.side_effect = get_enabled_services_side_effect
    get_port_details.side_effect = get_port_details_side_effect
    firewall = Firewall('test-firewall-1', ports=['test-port1', 'test-port2'],
                        is_external=False)
    results = firewall.diagnose()
    assert results == [
        DiagnosticCheck(
            'firewall-port-internal-test-port1',
            'Port {name} ({details}) available for internal '
            'networks', Result.PASSED, {
                'name': 'test-port1',
                'details': '1234/tcp, 1234/udp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-external-unavailable-test-port1',
            'Port {name} ({details}) unavailable for external '
            'networks', Result.PASSED, {
                'name': 'test-port1',
                'details': '1234/tcp, 1234/udp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-internal-test-port2',
            'Port {name} ({details}) available for internal networks',
            Result.FAILED, {
                'name': 'test-port2',
                'details': '2345/udp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-external-unavailable-test-port2',
            'Port {name} ({details}) unavailable for external networks',
            Result.FAILED, {
                'name': 'test-port2',
                'details': '2345/udp'
            }, 'test-firewall-1'),
    ]

    firewall = Firewall('test-firewall-1', ports=['test-port3', 'test-port4'],
                        is_external=True)
    results = firewall.diagnose()
    assert results == [
        DiagnosticCheck(
            'firewall-port-internal-test-port3',
            'Port {name} ({details}) available for internal networks',
            Result.PASSED, {
                'name': 'test-port3',
                'details': '3456/tcp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-external-available-test-port3',
            'Port {name} ({details}) available for external networks',
            Result.PASSED, {
                'name': 'test-port3',
                'details': '3456/tcp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-internal-test-port4',
            'Port {name} ({details}) available for internal networks',
            Result.FAILED, {
                'name': 'test-port4',
                'details': '4567/udp'
            }, 'test-firewall-1'),
        DiagnosticCheck(
            'firewall-port-external-available-test-port4',
            'Port {name} ({details}) available for external networks',
            Result.FAILED, {
                'name': 'test-port4',
                'details': '4567/udp'
            }, 'test-firewall-1'),
    ]


def test_local_protection_init():
    """Test initializing the local protection component."""
    component = FirewallLocalProtection('test-component', ['1234', '4567'])
    assert component.component_id == 'test-component'
    assert component.tcp_ports == ['1234', '4567']


@patch('plinth.modules.firewall.add_passthrough')
def test_local_protection_enable(add_passthrough):
    """Test enabling local protection component."""
    component = FirewallLocalProtection('test-component', ['1234', '4567'])
    component.enable()

    calls = [
        call('ipv6', '-A', 'INPUT', '-p', 'tcp', '--dport', '1234', '-j',
             'REJECT'),
        call('ipv4', '-A', 'INPUT', '-p', 'tcp', '--dport', '1234', '-j',
             'REJECT'),
        call('ipv6', '-A', 'INPUT', '-p', 'tcp', '--dport', '4567', '-j',
             'REJECT'),
        call('ipv4', '-A', 'INPUT', '-p', 'tcp', '--dport', '4567', '-j',
             'REJECT')
    ]
    add_passthrough.assert_has_calls(calls)


@patch('plinth.modules.firewall.remove_passthrough')
def test_local_protection_disable(remove_passthrough):
    """Test disabling local protection component."""
    component = FirewallLocalProtection('test-component', ['1234', '4567'])
    component.disable()

    calls = [
        call('ipv6', '-A', 'INPUT', '-p', 'tcp', '--dport', '1234', '-j',
             'REJECT'),
        call('ipv4', '-A', 'INPUT', '-p', 'tcp', '--dport', '1234', '-j',
             'REJECT'),
        call('ipv6', '-A', 'INPUT', '-p', 'tcp', '--dport', '4567', '-j',
             'REJECT'),
        call('ipv4', '-A', 'INPUT', '-p', 'tcp', '--dport', '4567', '-j',
             'REJECT')
    ]
    remove_passthrough.assert_has_calls(calls)


@patch('plinth.modules.firewall.components.FirewallLocalProtection.enable')
def test_local_protection_setup(enable):
    """Test setting up protection when updating the app."""

    class TestApp(App):
        app_id = 'test-app'
        enabled = True

        def is_enabled(self):
            return self.enabled

    app = TestApp()
    component = FirewallLocalProtection('test-component', ['1234', '4567'])
    app.add(component)

    component.setup(old_version=0)
    enable.assert_not_called()

    app.enabled = False
    component.setup(old_version=1)
    enable.assert_not_called()

    app.enabled = True
    component.setup(old_version=1)
    enable.assert_has_calls([call()])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for configuring Transmission Server.
"""

import logging
import socket

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views

from . import privileged
from .forms import TransmissionForm

logger = logging.getLogger(__name__)


class TransmissionAppView(views.AppView):
    """Serve configuration page."""
    form_class = TransmissionForm
    app_id = 'transmission'

    def get_initial(self):
        """Get the current settings from Transmission server."""
        status = super().get_initial()
        configuration = privileged.get_configuration()
        status['storage_path'] = configuration['download-dir']
        status['hostname'] = socket.gethostname()

        return status

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data
        if old_status['storage_path'] != new_status['storage_path']:
            new_configuration = {
                'download-dir': new_status['storage_path'],
            }
            privileged.merge_configuration(new_configuration)
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Gitweb."""

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import FormView

from plinth import app as app_module
from plinth import views
from plinth.modules import gitweb

from . import privileged
from .forms import CreateRepoForm, EditRepoForm


class GitwebAppView(views.AppView):
    """Serve configuration page."""

    app_id = 'gitweb'
    template_name = 'gitweb_configure.html'

    def get_context_data(self, *args, **kwargs):
        """Add repositories to the context data."""
        context = super().get_context_data(*args, **kwargs)
        repos = gitweb.get_repo_list()
        context['repos'] = repos
        context['cloning'] = any('clone_progress' in repo for repo in repos)
        if context['cloning']:
            context['refresh_page_sec'] = 3

        return context


class CreateRepoView(SuccessMessageMixin, FormView):
    """View to create a new repository."""

    form_class = CreateRepoForm
    prefix = 'gitweb'
    template_name = 'form.html'
    success_url = reverse_lazy('gitweb:index')
    success_message = _('Repository created.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Create Repository')
        return context

    def form_valid(self, form):
        """Create the repository on valid form submission."""
        form_data = {}
        for key, value in form.cleaned_data.items():
            if value is None:
                form_data[key] = ''
            else:
                form_data[key] = value
        try:
            gitweb.create_repo(form_data['name'], form_data['description'],
                               form_data['owner'], form_data['is_private'])
        except Exception as error:
            self.success_message = ''
            messages.error(
                self.request, "{0} {1}".format(
                    _('An error occurred while creating the repository.'),
                    error))
        else:
            app_module.App.get('gitweb').update_service_access()

        return super().form_valid(form)


class EditRepoView(SuccessMessageMixin, FormView):
    """View to edit an existing repository."""

    form_class = EditRepoForm
    prefix = 'gitweb'
    template_name = 'form.html'
    success_url = reverse_lazy('gitweb:index')
    success_message = _('Repository edited.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit repository')
        return context

    def get_initial(self):
        """Load information about repository being edited."""
        name = self.kwargs['name']
        for repo in gitweb.get_repo_list():
            if repo['name'] == name and 'clone_progress' not in repo:
                break
        else:
            raise Http404

        return gitweb.repo_info(name)

    def form_valid(self, form):
        """Edit the repo on valid form submission."""
        if form.initial != form.cleaned_data:
            form_data = {}
            for key, value in form.cleaned_data.items():
                if value is None:
                    form_data[key] = ''
                else:
                    form_data[key] = value

            try:
                gitweb.edit_repo(form.initial, form_data)
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))
        app_module.App.get('gitweb').update_service_access()

        return super().form_valid(form)


def delete(request, name):
    """Handle deleting repositories, showing a confirmation dialog first.

    On GET, display a confirmation page.
    On POST, delete the repository.
    """
    for repo in gitweb.get_repo_list():
        if repo['name'] == name and 'clone_progress' not in repo:
            break
    else:
        raise Http404

    app = app_module.App.get('gitweb')
    if request.method == 'POST':
        try:
            privileged.delete_repo(name)
            messages.success(request, _('{name} deleted.').format(name=name))
        except Exception as error:
            messages.error(
                request,
                _('Could not delete {name}: {error}').format(
                    name=name, error=error),
            )
        app.update_service_access()

        return redirect(reverse_lazy('gitweb:index'))

    return TemplateResponse(request, 'gitweb_delete.html', {
        'title': app.info.name,
        'name': name
    })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for upgrades."""

import subprocess

from apt.cache import Cache
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import TemplateView
from django.views.generic.edit import FormView

from plinth import __version__
from plinth.modules import first_boot, upgrades
from plinth.privileged import packages as packages_privileged
from plinth.views import AppView, messages_error

from . import privileged
from .forms import BackportsFirstbootForm, ConfigureForm, UpdateFirstbootForm


class UpgradesConfigurationView(AppView):
    """Serve configuration page."""

    form_class = ConfigureForm
    success_url = reverse_lazy('upgrades:index')
    template_name = "upgrades_configure.html"
    app_id = 'upgrades'

    def get_initial(self):
        """Return the initial values for the form."""
        return {
            'auto_upgrades_enabled': privileged.check_auto(),
            'dist_upgrade_enabled': upgrades.is_dist_upgrade_enabled()
        }

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['can_activate_backports'] = upgrades.can_activate_backports()
        context['is_backports_requested'] = upgrades.is_backports_requested()
        context['is_busy'] = (_is_updating()
                              or packages_privileged.is_package_manager_busy())
        context['log'] = privileged.get_log()
        context['refresh_page_sec'] = 3 if context['is_busy'] else None
        context['version'] = __version__
        context['new_version'] = is_newer_version_available()
        context['os_release'] = get_os_release()
        context['can_test_dist_upgrade'] = upgrades.can_test_dist_upgrade()
        return context

    def form_valid(self, form):
        """Apply the form changes."""
        old_status = form.initial
        new_status = form.cleaned_data

        is_changed = False

        if old_status['auto_upgrades_enabled'] \
           != new_status['auto_upgrades_enabled']:

            try:
                if new_status['auto_upgrades_enabled']:
                    privileged.enable_auto()
                else:
                    privileged.disable_auto()

                is_changed = True
            except Exception as exception:
                messages_error(self.request,
                               _('Error when configuring unattended-upgrades'),
                               exception)

        if old_status['dist_upgrade_enabled'] \
           != new_status['dist_upgrade_enabled']:
            upgrades.set_dist_upgrade_enabled(
                new_status['dist_upgrade_enabled'])
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated.'))

        return super().form_valid(form)


def is_newer_version_available():
    """Return whether a newer Freedombox version is available."""
    cache = Cache()
    freedombox = cache['freedombox']
    return not freedombox.candidate.is_installed


def get_os_release():
    """Return the Debian release number and name."""
    output = 'Error: Cannot read PRETTY_NAME in /etc/os-release.'
    with open('/etc/os-release', 'r', encoding='utf-8') as release_file:
        for line in release_file:
            if 'PRETTY_NAME=' in line:
                line = line.replace('"', '').strip()
                line = line.split('=')
                output = line[1]
    return output


def _is_updating():
    """Check if manually triggered update is running."""
    command = ['systemctl', 'is-active', 'freedombox-manual-upgrade']
    result = subprocess.run(command, capture_output=True, text=True,
                            check=False)
    return str(result.stdout).startswith('activ')  # 'active' or 'activating'


def upgrade(request):
    """Serve the upgrade page."""
    if request.method == 'POST':
        try:
            privileged.run()
            messages.success(request, _('Upgrade process started.'))
        except Exception:
            messages.error(request, _('Starting upgrade failed.'))

    return redirect(reverse_lazy('upgrades:index'))


def activate_backports(request):
    """Activate backports."""
    if request.method == 'POST':
        upgrades.set_backports_requested(True)
        upgrades.setup_repositories(None)
        messages.success(request, _('Frequent feature updates activated.'))

    return redirect(reverse_lazy('upgrades:index'))


class BackportsFirstbootView(FormView):
    """View to configure backports during first boot wizard."""

    template_name = 'backports-firstboot.html'
    form_class = BackportsFirstbootForm

    def dispatch(self, request, *args, **kwargs):
        """Show backports configuration form only if it can be activated."""
        if upgrades.is_backports_enabled():
            # Backports is already enabled. Record this preference and
            # skip first boot step.
            upgrades.set_backports_requested(True)
            first_boot.mark_step_done('backports_wizard')
            return HttpResponseRedirect(reverse_lazy(first_boot.next_step()))

        if not upgrades.can_activate_backports():
            # Skip first boot step.
            upgrades.set_backports_requested(False)
            first_boot.mark_step_done('backports_wizard')
            return HttpResponseRedirect(reverse_lazy(first_boot.next_step()))

        return super().dispatch(request, *args, *kwargs)

    def get_success_url(self):
        """Return next firstboot step."""
        return reverse_lazy(first_boot.next_step())

    def form_valid(self, form):
        """Mark the first wizard step as done, save value and redirect."""
        enabled = form.cleaned_data['backports_enabled']
        upgrades.set_backports_requested(enabled)
        upgrades.setup_repositories(None)
        first_boot.mark_step_done('backports_wizard')
        return super().form_valid(form)


class UpdateFirstbootView(FormView):
    """View to run initial update during first boot wizard."""

    template_name = 'update-firstboot.html'
    form_class = UpdateFirstbootForm

    def __init__(self):
        """Define instance attribute."""
        self.update = True

    def get_success_url(self):
        """Return next firstboot step."""
        if self.update:
            return reverse_lazy('upgrades:update-firstboot-progress')

        return reverse_lazy(first_boot.next_step())

    def form_valid(self, form):
        """Run update if selected, and mark step as done."""
        self.update = form.cleaned_data['update_now']
        if self.update:
            privileged.run()

        first_boot.mark_step_done('initial_update')
        return super().form_valid(form)


class UpdateFirstbootProgressView(TemplateView):
    """View to show initial update progress."""

    template_name = 'update-firstboot-progress.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['is_busy'] = (_is_updating()
                              or packages_privileged.is_package_manager_busy())
        context['next_step'] = first_boot.next_step()
        context['refresh_page_sec'] = 3 if context['is_busy'] else None
        return context


def test_dist_upgrade(request):
    """Test dist-upgrade from stable to testing."""
    if request.method == 'POST':
        upgrades.test_dist_upgrade()
        messages.success(request, _('Starting distribution upgrade test.'))

    return redirect(reverse_lazy('upgrades:index'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Various helpers for the I2P app.
"""

import os
import re
from collections import OrderedDict

import augeas

I2P_CONF_DIR = '/var/lib/i2p/i2p-config'
FILE_TUNNEL_CONF = os.path.join(I2P_CONF_DIR, 'i2ptunnel.config')
TUNNEL_IDX_REGEX = re.compile(r'tunnel.(\d+).name$')
I2P_ROUTER_CONF = os.path.join(I2P_CONF_DIR, 'router.config')


class TunnelEditor():
    """Helper to edit I2P tunnel configuration file using augeas.

    :type aug: augeas.Augeas

    """
    def __init__(self, conf_filename=None, idx=None):
        self.conf_filename = conf_filename or FILE_TUNNEL_CONF
        self.idx = idx
        self.aug = None

    @property
    def lines(self):
        """Return lines from configuration file."""
        if self.aug:
            return self.aug.match('/files{}/*'.format(self.conf_filename))

        return []

    def read_conf(self):
        """Load an instance of Augeaus for processing APT configuration.

        Chainable method.

        """
        self.aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                                 augeas.Augeas.NO_MODL_AUTOLOAD)
        self.aug.set('/augeas/load/Properties/lens', 'Properties.lns')
        self.aug.set('/augeas/load/Properties/incl[last() + 1]',
                     self.conf_filename)
        self.aug.load()

        return self

    def write_conf(self):
        """Write changes to the configuration file to disk.

        Chainable method.

        """
        self.aug.save()
        return self

    def set_tunnel_idx(self, name):
        """Finds the index of the tunnel with the given name.

        Chainable method.

        :type name: basestring

        """
        for prop in self.aug.match('/files{}/*'.format(self.conf_filename)):
            match = TUNNEL_IDX_REGEX.search(prop)
            if match and self.aug.get(prop) == name:
                self.idx = int(match.group(1))
                return self

        raise ValueError('No tunnel called {}'.format(name))

    def calc_prop_path(self, tunnel_prop):
        """Calculates the property name as found in the properties files.

        :type tunnel_prop: str
        :rtype: basestring

        """
        calced_prop_path = \
            '/files{filepath}/tunnel.{idx}.{tunnel_prop}'.format(
                idx=self.idx,
                tunnel_prop=tunnel_prop,
                filepath=self.conf_filename)
        return calced_prop_path

    def set_tunnel_prop(self, tunnel_prop, value):
        """Updates a tunnel's property.

        The idx has to be set and the property has to exist in the config file
        and belong to the tunnel's properties.

        See calc_prop_path.

        Chainable method.

        :param tunnel_prop:
        :type tunnel_prop: str
        :param value:
        :type value: basestring | int
        :return:
        :rtype:

        """
        if self.idx is None:
            raise ValueError(
                'Please init the tunnel index before calling this method')

        calc_prop_path = self.calc_prop_path(tunnel_prop)
        self.aug.set(calc_prop_path, value)
        return self

    def __getitem__(self, tunnel_prop):
        ret = self.aug.get(self.calc_prop_path(tunnel_prop))
        if ret is None:
            raise KeyError('Unknown property {}'.format(tunnel_prop))

        return ret

    def __setitem__(self, tunnel_prop, value):
        self.aug.set(self.calc_prop_path(tunnel_prop), value)


class RouterEditor():
    """Helper to edit I2P router configuration file using augeas.

    :type aug: augeas.Augeas

    """

    FAVORITE_PROP = 'routerconsole.favorites'
    FAVORITE_TUPLE_SIZE = 4

    def __init__(self, filename=None):
        self.conf_filename = filename or I2P_ROUTER_CONF
        self.aug = None

    def read_conf(self):
        """Load an instance of Augeaus for processing APT configuration.

        Chainable method.

        """
        self.aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                                 augeas.Augeas.NO_MODL_AUTOLOAD)
        self.aug.set('/augeas/load/Properties/lens', 'Properties.lns')
        self.aug.set('/augeas/load/Properties/incl[last() + 1]',
                     self.conf_filename)
        self.aug.load()
        return self

    def write_conf(self):
        """Write changes to the configuration file to disk.

        Chainable method.

        """
        self.aug.save()
        return self

    @property
    def favorite_property(self):
        """Return the favourites property from configuration file."""
        return '/files{filename}/{prop}'.format(filename=self.conf_filename,
                                                prop=self.FAVORITE_PROP)

    def add_favorite(self, name, url, description=None, icon=None):
        """Add a favorite to the router configuration file.

        Favorites are in a single string and separated by ','. none of the
        incoming params can therefore use commas. I2P replaces the commas by
        dots.

        That's ok for the name and description, but not for the url and icon.

        :type name: basestring
        :type url: basestring
        :type description: basestring
        :type icon: basestring

        """
        if not description:
            description = ''

        if not icon:
            icon = '/themes/console/images/eepsite.png'

        if ',' in url:
            raise ValueError('URL cannot contain commas')

        if ',' in icon:
            raise ValueError('Icon cannot contain commas')

        name = name.replace(',', '.')
        description = description.replace(',', '.')

        prop = self.favorite_property
        favorites = self.aug.get(prop) or ''
        new_favorite = '{name},{description},{url},{icon},'.format(
            name=name, description=description, url=url, icon=icon)
        self.aug.set(prop, favorites + new_favorite)
        return self

    def get_favorites(self):
        """Return list of favorites."""
        favs_string = self.aug.get(self.favorite_property) or ''
        favs_split = favs_string.split(',')

        # There's a trailing comma --> 1 extra
        favs_len = len(favs_split)
        if favs_len > 0:
            favs_split = favs_split[:-1]
            favs_len = len(favs_split)

        if favs_len % self.FAVORITE_TUPLE_SIZE:
            raise SyntaxError("Invalid number of fields in favorite line")

        favs = OrderedDict()
        for index in range(0, favs_len, self.FAVORITE_TUPLE_SIZE):
            next_index = index + self.FAVORITE_TUPLE_SIZE
            name, description, url, icon = favs_split[index:next_index]
            favs[url] = {
                'name': name,
                'description': description,
                'icon': icon
            }

        return favs



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure I2P."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.i2p.resources import FAVORITES
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages

from . import manifest, privileged

_description = [
    _('The Invisible Internet Project is an anonymous network layer intended '
      'to protect communication from censorship and surveillance. I2P '
      'provides anonymity by sending encrypted traffic through a '
      'volunteer-run network distributed around the world.'),
    _('Find more information about I2P on their project '
      '<a href="https://geti2p.net" target="_blank">homepage</a>.'),
    _('The first visit to the provided web interface will initiate the '
      'configuration process.')
]

tunnels_to_manage = {
    'I2P HTTP Proxy': 'i2p-http-proxy-freedombox',
    'I2P HTTPS Proxy': 'i2p-https-proxy-freedombox',
    'Irc2P': 'i2p-irc-freedombox'
}


class I2PApp(app_module.App):
    """FreedomBox app for I2P."""

    app_id = 'i2p'

    _version = 3

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {'i2p': _('Manage I2P application')}

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('I2P'), icon_filename='i2p',
                               short_description=_('Anonymity Network'),
                               description=_description, manual_page='I2P',
                               clients=manifest.clients)
        self.add(info)

        menu_item = menu.Menu('menu-i2p', info.name, info.short_description,
                              info.icon_filename, 'i2p:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-i2p', info.name,
                                      short_description=info.short_description,
                                      icon=info.icon_filename, url='/i2p/',
                                      clients=info.clients,
                                      login_required=True,
                                      allowed_groups=list(groups))
        self.add(shortcut)

        packages = Packages('packages-i2p', ['i2p'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-i2p', [
            '/etc/apache2/conf-available/i2p-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-i2p-web', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall = Firewall('firewall-i2p-proxies', _('I2P Proxy'),
                            ports=tunnels_to_manage.values(),
                            is_external=False)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-i2p', ['7657'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-i2p', 'i2p-freedombox',
                              urls=['https://{host}/i2p/'])
        self.add(webserver)

        daemon = Daemon('daemon-i2p', 'i2p', listen_ports=[(7657, 'tcp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-i2p',
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-i2p', **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        if not old_version:
            self.disable()
            # Add favorites to the configuration
            for fav in FAVORITES:
                privileged.add_favorite(fav['name'], fav['url'],
                                        fav.get('description'),
                                        fav.get('icon'))

            # Tunnels to all interfaces
            for tunnel in tunnels_to_manage:
                privileged.set_tunnel_property(tunnel, 'interface', '0.0.0.0')

            self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Nextcloud."""

import contextlib
import json
import pathlib
import secrets
import shutil
import string
import subprocess
import time

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONTAINER_NAME = 'nextcloud-freedombox'
SERVICE_NAME = 'nextcloud-freedombox'
VOLUME_NAME = 'nextcloud-freedombox'
IMAGE_NAME = 'docker.io/library/nextcloud:stable-fpm'

DB_HOST = 'localhost'
DB_NAME = 'nextcloud_fbx'
DB_USER = 'nextcloud_fbx'
GUI_ADMIN = 'nextcloud-admin'
REDIS_DB = 8  # Don't clash with other redis apps

_data_path = pathlib.Path('/var/lib/nextcloud/')

DB_BACKUP_FILE = pathlib.Path(
    '/var/lib/plinth/backups-data/nextcloud-database.sql')


@privileged
def setup():
    """Setup Nextcloud configuration."""
    # Setup redis for caching
    _redis_listen_socket()

    volumes = {
        '/run/mysqld/mysqld.sock': '/run/mysqld/mysqld.sock',
        '/run/redis/redis-server.sock': '/run/redis/redis-server.sock',
        '/run/slapd/ldapi': '/run/slapd/ldapi',
        VOLUME_NAME: '/var/www/html'
    }
    env = {'OVERWRITEWEBROOT': '/nextcloud'}
    binds_to = ['mariadb.service', 'redis-server.service', 'slapd.service']
    action_utils.podman_create(container_name=CONTAINER_NAME,
                               image_name=IMAGE_NAME, volume_name=VOLUME_NAME,
                               volume_path=str(_data_path), volumes=volumes,
                               env=env, binds_to=binds_to)
    action_utils.service_start(CONTAINER_NAME)

    _nextcloud_wait_until_ready()

    # Setup database
    _create_database()
    database_password = _get_database_password()
    if not database_password:
        database_password = _generate_secret_key(16)
        _set_database_privileges(database_password)

    # Setup redis configuration
    _create_redis_config()

    # Run setup wizard
    _nextcloud_setup_wizard(database_password)

    # Setup LDAP configuraiton
    _configure_ldap()


def _run_in_container(
        *args, capture_output: bool = False, check: bool = True,
        env: dict[str, str] | None = None) -> subprocess.CompletedProcess:
    """Run a command inside the container."""
    env_args = [f'--env={key}={value}' for key, value in (env or {}).items()]
    command = ['podman', 'exec', '--user', 'www-data'
               ] + env_args + [CONTAINER_NAME] + list(args)
    return subprocess.run(command, capture_output=capture_output, check=check)


def _run_occ(*args, **kwargs) -> subprocess.CompletedProcess:
    """Run the Nextcloud occ command inside the container."""
    return _run_in_container('/var/www/html/occ', *args, **kwargs)


@privileged
def is_enabled() -> bool:
    """Return if the systemd container service is enabled."""
    return action_utils.podman_is_enabled(CONTAINER_NAME)


@privileged
def enable():
    """Enable the systemd container service."""
    action_utils.podman_enable(CONTAINER_NAME)


@privileged
def disable():
    """Disable the systemd container service."""
    action_utils.podman_disable(CONTAINER_NAME)


@privileged
def get_override_domain():
    """Return the domain name that Nextcloud is configured to override with."""
    try:
        domain = _run_occ('config:system:get', 'overwritehost',
                          capture_output=True)
        return domain.stdout.decode().strip()
    except subprocess.CalledProcessError:
        return None


@privileged
def set_override_domain(domain_name: str):
    """Set the domain name that Nextcloud will use to override all domains."""
    protocol = 'https'
    if domain_name.endswith('.onion'):
        protocol = 'http'

    if domain_name:
        _run_occ('config:system:set', 'overwritehost', '--value', domain_name)
        _run_occ('config:system:set', 'overwriteprotocol', '--value', protocol)
        _run_occ('config:system:set', 'overwrite.cli.url', '--value',
                 f'{protocol}://{domain_name}/nextcloud')
    else:
        _run_occ('config:system:delete', 'overwritehost')
        _run_occ('config:system:delete', 'overwriteprotocol')
        _run_occ('config:system:delete', 'overwrite.cli.url')

    # Restart to apply changes immediately
    action_utils.service_restart('nextcloud-freedombox')


@privileged
def set_trusted_domains(domains: list[str]):
    """Set the list of trusted domains."""
    _run_occ('config:system:delete', 'trusted_domains')
    for index, domain in enumerate(domains):
        _run_occ('config:system:set', 'trusted_domains', str(index), '--value',
                 domain)


@privileged
def set_admin_password(password: str):
    """Set password for owncloud-admin"""
    _run_occ('user:resetpassword', '--password-from-env', GUI_ADMIN,
             env={'OC_PASS': password})


@privileged
def get_default_phone_region():
    """"Get the value of default_phone_region."""
    try:
        default_phone_region = _run_occ('config:system:get',
                                        'default_phone_region',
                                        capture_output=True)
        return default_phone_region.stdout.decode().strip()
    except subprocess.CalledProcessError:
        return None


@privileged
def set_default_phone_region(region: str):
    """"Set the value of default_phone_region."""
    _run_occ('config:system:set', 'default_phone_region', '--value', region)


def _database_query(query: str):
    """Run a database query."""
    subprocess.run(['mysql'], input=query.encode(), check=True)


def _create_database():
    """Create an empty MySQL database for Nextcloud."""
    # SQL injection is avoided due to known input.
    _db_file_path = pathlib.Path('/var/lib/mysql/nextcloud_fbx')
    if _db_file_path.exists():
        return

    query = f'CREATE DATABASE IF NOT EXISTS {DB_NAME} ' \
        'CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci;'
    _database_query(query)


def _set_database_privileges(db_password: str):
    """Create user, set password and provide permissions on the database."""
    queries = [
        f"CREATE USER IF NOT EXISTS '{DB_USER}'@'localhost';",
        f"GRANT ALL PRIVILEGES ON {DB_NAME}.* TO '{DB_USER}'@'localhost';",
        f"ALTER USER '{DB_USER}'@'localhost' IDENTIFIED BY '{db_password}';",
    ]
    for query in queries:
        _database_query(query)


def _nextcloud_wait_until_ready():
    """Wait for Nextcloud container to get ready."""
    # Nextcloud copies sources from /usr/src/nextcloud to /var/www/html inside
    # the container. Nextcloud is served from the latter location. This happens
    # on first run of the container and when upgrade happen.
    start_time = time.time()
    while time.time() < start_time + 300:
        if (_data_path / 'version.php').exists():
            break

        time.sleep(1)

    # Wait while Nextcloud is syncing files, running install or performing an
    # upgrade by trying to obtain an exclusive on its init-sync.lock. Wrap the
    # echo command with the lock so that the lock is immediately released after
    # obtaining. We are unable to obtain the lock for 5 minutes, fail and stop
    # the setup process.
    lock_file = _data_path / 'nextcloud-init-sync.lock'
    subprocess.run(
        ['flock', '--exclusive', '--wait', '300', lock_file, 'echo'],
        check=True)


def _nextcloud_get_status():
    """Return Nextcloud status such installed, in maintenance, etc."""
    output = _run_occ('status', '--output=json', capture_output=True)
    return json.loads(output.stdout)


def _nextcloud_setup_wizard(db_password: str):
    """Run the Nextcloud installation wizard and enable cron jobs."""
    if not _nextcloud_get_status()['installed']:
        admin_password = _generate_secret_key(16)
        _run_occ('maintenance:install', '--database=mysql',
                 '--database-host=localhost:/run/mysqld/mysqld.sock',
                 f'--database-name={DB_NAME}', f'--database-user={DB_USER}',
                 f'--database-pass={db_password}', f'--admin-user={GUI_ADMIN}',
                 f'--admin-pass={admin_password}')

    # For the server to work properly, it's important to configure background
    # jobs correctly. Cron is the recommended setting.
    _run_occ('background:cron')

    # Enable pretty URLs without /index.php in them.
    _run_occ('config:system:set', 'htaccess.RewriteBase', '--value',
             '/nextcloud')
    _run_occ('config:system:set', 'htaccess.IgnoreFrontController',
             '--type=boolean', '--value=true')
    # Update the .htaccess file to contain mod_rewrite rules needed for pretty
    # URLs. This is automatically re-run by scripts when upgrading to next
    # version.
    _run_occ('maintenance:update:htaccess')


def _configure_ldap():
    _run_occ('app:enable', 'user_ldap')

    # Check if LDAP has already been configured. This is necessary because
    # if the setup proccess is rerun when updating the FredomBox app another
    # redundant LDAP config would be created.
    output = _run_occ('ldap:test-config', 's01', capture_output=True,
                      check=False)
    if 'Invalid configID' in output.stdout.decode():
        _run_occ('ldap:create-empty-config')

    ldap_settings = {
        'ldapBase': 'dc=thisbox',
        'ldapBaseGroups': 'dc=thisbox',
        'ldapBaseUsers': 'dc=thisbox',
        'ldapConfigurationActive': '1',
        'ldapGroupDisplayName': 'cn',
        'ldapGroupFilter': '(&(|(objectclass=posixGroup)))',
        'ldapGroupFilterMode': '0',
        'ldapGroupFilterObjectclass': 'posixGroup',
        'ldapGroupMemberAssocAttr': 'memberUid',
        'ldapHost': 'ldapi:///',
        'ldapLoginFilter': '(&(|(objectclass=posixAccount))(uid=%uid))',
        'ldapLoginFilterEmail': '0',
        'ldapLoginFilterMode': '0',
        'ldapLoginFilterUsername': '1',
        'ldapNestedGroups': '0',
        'ldapUserDisplayName': 'cn',
        'ldapUserFilter': '(|(objectclass=posixAccount))',
        'ldapUserFilterMode': '0',
        'ldapUserFilterObjectclass': 'account',
        'ldapUuidGroupAttribute': 'auto',
        'ldapUuidUserAttribute': 'auto',
        'turnOffCertCheck': '0',
        'turnOnPasswordChange': '0',
        'useMemberOfToDetectMembership': '0'
    }

    for key, value in ldap_settings.items():
        _run_occ('ldap:set-config', 's01', key, value)


@privileged
def uninstall():
    """Uninstall Nextcloud"""
    _drop_database()
    action_utils.podman_uninstall(container_name=CONTAINER_NAME,
                                  volume_name=VOLUME_NAME,
                                  image_name=IMAGE_NAME,
                                  volume_path=str(_data_path))


def _drop_database():
    """Drop the database that was created during install."""
    with action_utils.service_ensure_running('mysql'):
        _database_query(f'DROP DATABASE IF EXISTS {DB_NAME};')
        _database_query(f"DROP USER IF EXISTS '{DB_USER}'@'localhost';")


def _generate_secret_key(length=64, chars=None):
    """Generate a new random secret key for use with Nextcloud."""
    chars = chars or (string.ascii_letters + string.digits)
    return ''.join(secrets.choice(chars) for _ in range(length))


def _set_maintenance_mode(on: bool):
    """Turn maintenance mode on or off."""
    _run_occ('maintenance:mode', '--on' if on else '--off')


@contextlib.contextmanager
def _maintenance_mode():
    """Context to set maintenance mode temporarily."""
    try:
        _set_maintenance_mode(True)
        yield
    finally:
        _set_maintenance_mode(False)


@privileged
def dump_database():
    """Dump database to file."""
    DB_BACKUP_FILE.parent.mkdir(parents=True, exist_ok=True)

    with _maintenance_mode():
        with DB_BACKUP_FILE.open('w', encoding='utf-8') as file_handle:
            subprocess.run([
                'mysqldump', '--add-drop-database', '--add-drop-table',
                '--add-drop-trigger', '--single-transaction',
                '--default-character-set=utf8mb4', '--user', 'root',
                '--databases', DB_NAME
            ], stdout=file_handle, check=True)


@privileged
def restore_database():
    """Restore database from file."""
    with DB_BACKUP_FILE.open('r', encoding='utf-8') as file_handle:
        subprocess.run(['mysql', '--user', 'root'], stdin=file_handle,
                       check=True)

    subprocess.run(['redis-cli', '-n',
                    str(REDIS_DB), 'FLUSHDB', 'SYNC'], check=False)

    _set_database_privileges(_get_database_password())

    # After updating the configuration, a restart seems to be required for the
    # new DB password be used.
    action_utils.service_try_restart(SERVICE_NAME)

    _set_maintenance_mode(False)

    # Attempts to update UUIDs of user and group entries. By default, the
    # command attempts to update UUIDs that have been invalidated by a
    # migration step.
    _run_occ('ldap:update-uuid')

    # Update the systems data-fingerprint after a backup is restored.
    _run_occ('maintenance:data-fingerprint')


def _get_database_password():
    """Return the database password from config.php or '' if not set.

    OCC cannot run unless Nextcloud can already connect to the database.
    """
    code = 'if (file_exists("/var/www/html/config/config.php")) {' \
        'include_once("/var/www/html/config/config.php");' \
        'print($CONFIG["dbpassword"] ?? ""); }'
    return _run_in_container('php', '-r', code,
                             capture_output=True).stdout.decode().strip()


def _create_redis_config():
    """Create a php file for Redis configuration."""
    config_file = _data_path / 'config/freedombox.config.php'
    file_content = fr'''<?php
$CONFIG = [
'memcache.distributed' => '\OC\Memcache\Redis',
'memcache.locking' => '\OC\Memcache\Redis',
'redis' => ['host' => '/run/redis/redis-server.sock', 'dbindex' => {REDIS_DB}],
];
'''
    config_file.write_text(file_content)
    shutil.chown(config_file, 'www-data', 'www-data')


def _load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    redis_config = '/etc/redis/redis.conf'
    aug.transform('Spacevars', redis_config)
    aug.set('/augeas/context', '/files' + redis_config)
    aug.load()
    return aug


def _redis_listen_socket():
    """Configure Redis to listen on a UNIX socket."""
    aug = _load_augeas()
    value = '/etc/redis/conf.d/*.conf'
    found = any((aug.get(match_) == value for match_ in aug.match('include')))
    if not found:
        aug.set('include[last() + 1]', value)
        aug.save()
        action_utils.service_restart('redis-server')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Tiny Tiny RSS."""

import os
import subprocess

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONFIG_FILE = '/etc/tt-rss/config.php'
DEFAULT_FILE = '/etc/default/tt-rss'
DATABASE_FILE = '/etc/tt-rss/database.php'
DB_BACKUP_FILE = '/var/lib/plinth/backups-data/ttrss-database.sql'


@privileged
def pre_setup():
    """Preseed debconf values before packages are installed."""
    action_utils.debconf_set_selections([
        'tt-rss tt-rss/database-type string pgsql',
        'tt-rss tt-rss/purge boolean true',
        'tt-rss tt-rss/dbconfig-remove boolean true',
        'tt-rss tt-rss/dbconfig-reinstall boolean true'
    ])


@privileged
def get_domain() -> str | None:
    """Get the domain set for Tiny Tiny RSS."""
    aug = load_augeas()

    from urllib.parse import urlparse
    for match in aug.match('/files' + CONFIG_FILE + '/define'):
        if aug.get(match) == 'SELF_URL_PATH':
            url = aug.get(match + '/value').strip("'")
            return urlparse(url).netloc

    return None


@privileged
def set_domain(domain_name: str | None):
    """Set the domain to be used by Tiny Tiny RSS."""
    if not domain_name:
        return

    url = f"'https://{domain_name}/tt-rss/'"
    aug = load_augeas()

    for match in aug.match('/files' + CONFIG_FILE + '/define'):
        if aug.get(match) == 'SELF_URL_PATH':
            aug.set(match + '/value', url)

    aug.save()


@privileged
def setup():
    """Setup Tiny Tiny RSS configuration."""
    aug = load_augeas()

    aug.set('/files' + DEFAULT_FILE + '/DISABLED', '0')

    skip_self_url_path_exists = False

    for match in aug.match('/files' + CONFIG_FILE + '/define'):
        if aug.get(match) == 'PLUGINS':
            aug.set(match + '/value', "'auth_remote, note'")
        elif aug.get(match) == '_SKIP_SELF_URL_PATH_CHECKS':
            skip_self_url_path_exists = True
            aug.set(match + '/value', 'true')

    if not skip_self_url_path_exists:
        aug.set('/files' + CONFIG_FILE + '/define[last() + 1]',
                '_SKIP_SELF_URL_PATH_CHECKS')
        aug.set('/files' + CONFIG_FILE + '/define[last()]/value', 'true')

    aug.save()

    if action_utils.service_is_enabled('tt-rss'):
        action_utils.service_restart('tt-rss')


def _get_database_config():
    """Return the database configuration."""
    aug = load_augeas()

    def _get_value(variable_name):
        """Return the value of a variable from database configuration file."""
        return aug.get('/files' + DATABASE_FILE + '/$' + variable_name) \
            .strip("'\"")

    return {
        'user': _get_value('dbuser'),
        'password': _get_value('dbpass'),
        'database': _get_value('dbname'),
        'host': _get_value('dbserver')
    }


@privileged
def enable_api_access():
    """Enable API access so that tt-rss can be accessed through mobile app."""
    import psycopg2  # Only available post installation

    config = _get_database_config()

    connection = psycopg2.connect(database=config['database'],
                                  user=config['user'],
                                  password=config['password'],
                                  host=config['host'])
    cursor = connection.cursor()

    cursor.execute("UPDATE ttrss_prefs SET def_value=true "
                   "WHERE pref_name='ENABLE_API_ACCESS';")

    connection.commit()
    connection.close()


@privileged
def dump_database():
    """Dump database to file."""
    config = _get_database_config()
    os.makedirs(os.path.dirname(DB_BACKUP_FILE), exist_ok=True)
    with open(DB_BACKUP_FILE, 'w', encoding='utf-8') as db_backup_file:
        process = _run_as_postgres(['pg_dumpall', '--roles-only'],
                                   stdout=subprocess.PIPE)
        db_backup_file.write(f'DROP ROLE IF EXISTS {config["user"]};\n')
        for line in process.stdout.decode().splitlines():
            if config['user'] in line:
                db_backup_file.write(line + '\n')

    with open(DB_BACKUP_FILE, 'a', encoding='utf-8') as db_backup_file:
        _run_as_postgres([
            'pg_dump', '--create', '--clean', '--if-exists', config['database']
        ], stdout=db_backup_file)


@privileged
def restore_database():
    """Restore database from file."""
    config = _get_database_config()

    # This is needed for old backups only. New backups include 'DROP DATABASE
    # IF EXISTS' and 'CREATE DATABASE' statements.
    _run_as_postgres(['dropdb', config['database']])
    _run_as_postgres(['createdb', config['database']])

    with open(DB_BACKUP_FILE, 'r', encoding='utf-8') as db_restore_file:
        _run_as_postgres(['psql', '--dbname', config['database']],
                         stdin=db_restore_file)


def _run_as_postgres(command, stdin=None, stdout=None):
    """Run a command as postgres user."""
    command = ['sudo', '--user', 'postgres'] + command
    return subprocess.run(command, stdin=stdin, stdout=stdout, check=True)


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Shellvars/lens', 'Shellvars.lns')
    aug.set('/augeas/load/Shellvars/incl[last() + 1]', DEFAULT_FILE)
    aug.set('/augeas/load/Phpvars/lens', 'Phpvars.lns')
    aug.set('/augeas/load/Phpvars/incl[last() + 1]', CONFIG_FILE)
    aug.set('/augeas/load/Phpvars/incl[last() + 1]', DATABASE_FILE)
    aug.load()
    return aug


@privileged
def uninstall():
    """Ensure that the database is removed."""
    # This setting set before deb package installation is not retrained,
    # somehow.
    action_utils.debconf_set_selections(['tt-rss tt-rss/purge boolean true'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for Version class.
"""

from plinth.version import Version


def test_version_comparisons():
    """Test comparing Debian package version numbers."""
    assert Version('3.1.8-1') == Version('3.1.8-1')
    assert Version('3.1.8-1') <= Version('4~')
    assert Version('3.1.8-1') < Version('4~')
    assert Version('4.0.0-1') >= Version('4~')
    assert Version('4.0') >= Version('4~')
    assert Version('4.0.0-1') > Version('4~')
    assert Version('4.0') > Version('4~')


def test_backport_versions():
    """Test comparing Debian backports package version numbers."""
    assert Version('3.1.7-1~bpo11+1') == Version('3.1.7-1~bpo11+1')
    assert Version('3.1.7-1~bpo11+1') <= Version('4~')
    assert Version('3.1.7-1~bpo11+1') < Version('4~')
    assert Version('4.0.0-1~bpo11+1') >= Version('4~')
    assert Version('4.0.0-1~bpo11+1') > Version('4~')


def test_dfsg_versions():
    """Test comparing Debian DFSG package version numbers."""
    assert Version('1.3.0+dfsg-2.2') == Version('1.3.0+dfsg-2.2')
    assert Version('1.3.0+dfsg-2.2') <= Version('1.4~')
    assert Version('1.3.0+dfsg-2.2') < Version('1.4~')
    assert Version('1.4.0+dfsg-1.1') >= Version('1.4~')
    assert Version('1.4.0+dfsg-1.1') > Version('1.4~')


def test_git_versions():
    """Test comparing Debian git package version numbers."""
    assert Version('21~git20210204.b4cbc79+dfsg-1') == \
        Version('21~git20210204.b4cbc79+dfsg-1')
    assert Version('21~git20210204.b4cbc79+dfsg-1') <= Version('22~')
    assert Version('21~git20210204.b4cbc79+dfsg-1') < Version('22~')
    assert Version('22~git20210204.b4cbc79+dfsg-1') >= Version('22~')
    assert Version('22~git20210204.b4cbc79+dfsg-1') > Version('22~')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for setup module.
"""

from plinth.setup import store_error_message, retrieve_error_messages


def test_store_retrieve_error_message():
    """Test storing and retrieving error messages."""
    store_error_message('error 1')
    assert retrieve_error_messages() == ['error 1']

    store_error_message('error 1')
    store_error_message('error 2')
    assert retrieve_error_messages() == ['error 1', 'error 2']

    # errors are cleared after retrieving
    assert retrieve_error_messages() == []



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for Samba.
"""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

SHARES_CONF_BACKUP_FILE = '/var/lib/plinth/backups-data/samba-shares-dump.conf'

clients = [{
    'name':
        _('Android Samba Client'),
    'platforms': [{
        'type':
            'store',
        'os':
            'android',
        'store_name':
            'f-droid',
        'url':
            store_url('f-droid', 'com.google.android.sambadocumentsprovider')
    }]
}, {
    'name':
        _('Ghost Commander'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'com.ghostsq.commander')
    }]
}, {
    'name':
        _('Ghost Commander - Samba plugin'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'com.ghostsq.commander.smb')
    }]
}, {
    'name':
        _('VLC media player'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'org.videolan.vlc')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'org.videolan.vlc')
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/app/apple-store/id650377962'
    }]
}, {
    'name':
        _('GNOME Files'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'nautilus',
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://wiki.gnome.org/Apps/Files'
    }]
}, {
    'name':
        _('Dolphin'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'dolphin',
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://kde.org/applications/system/org.kde.dolphin'
    }]
}]

backup = {'data': {'files': [SHARES_CONF_BACKUP_FILE]}, 'services': ['smbd']}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the upgrades module
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/upgrades/$', views.UpgradesConfigurationView.as_view(),
            name='index'),
    re_path(r'^sys/upgrades/activate-backports/$', views.activate_backports,
            name='activate-backports'),
    re_path(r'^sys/upgrades/firstboot/backports/$',
            views.BackportsFirstbootView.as_view(),
            name='backports-firstboot'),
    re_path(r'^sys/upgrades/firstboot/update/$',
            views.UpdateFirstbootView.as_view(), name='update-firstboot'),
    re_path(r'^sys/upgrades/firstboot/update/progress/$',
            views.UpdateFirstbootProgressView.as_view(),
            name='update-firstboot-progress'),
    re_path(r'^sys/upgrades/upgrade/$', views.upgrade, name='upgrade'),
    re_path(r'^sys/upgrades/test-dist-upgrade/$', views.test_dist_upgrade,
            name='test-dist-upgrade'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-02-10 12:30
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('plinth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Module',
            fields=[
                ('name', models.TextField(primary_key=True, serialize=False)),
                ('setup_version', models.IntegerField()),
            ],
        ),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the backups module.
"""

from django.urls import re_path

from .views import (AddRemoteRepositoryView, AddRepositoryView, BackupsView,
                    CreateArchiveView, DeleteArchiveView, DownloadArchiveView,
                    RemoveRepositoryView, RestoreArchiveView,
                    RestoreFromUploadView, ScheduleView, UploadArchiveView,
                    VerifySshHostkeyView, mount_repository, umount_repository)

urlpatterns = [
    re_path(r'^sys/backups/$', BackupsView.as_view(), name='index'),
    re_path(r'^sys/backups/(?P<uuid>[^/]+)/schedule/$', ScheduleView.as_view(),
            name='schedule'),
    re_path(r'^sys/backups/create/(?:(?P<app_id>[1-9a-z\-_]+)/)?$',
            CreateArchiveView.as_view(), name='create'),
    re_path(r'^sys/backups/(?P<uuid>[^/]+)/download/(?P<name>[^/]+)/$',
            DownloadArchiveView.as_view(), name='download'),
    re_path(r'^sys/backups/(?P<uuid>[^/]+)/delete/(?P<name>[^/]+)/$',
            DeleteArchiveView.as_view(), name='delete'),
    re_path(r'^sys/backups/upload/$', UploadArchiveView.as_view(),
            name='upload'),
    re_path(r'^sys/backups/(?P<uuid>[^/]+)/restore-archive/(?P<name>[^/]+)/$',
            RestoreArchiveView.as_view(), name='restore-archive'),
    re_path(r'^sys/backups/restore-from-upload/$',
            RestoreFromUploadView.as_view(), name='restore-from-upload'),
    re_path(r'^sys/backups/repositories/add/$', AddRepositoryView.as_view(),
            name='add-repository'),
    re_path(r'^sys/backups/repositories/add-remote/$',
            AddRemoteRepositoryView.as_view(), name='add-remote-repository'),
    re_path(r'^sys/backups/repositories/(?P<uuid>[^/]+)/ssh-verify/$',
            VerifySshHostkeyView.as_view(), name='verify-ssh-hostkey'),
    re_path(r'^sys/backups/repositories/(?P<uuid>[^/]+)/delete/$',
            RemoveRepositoryView.as_view(), name='repository-remove'),
    re_path(r'^sys/backups/repositories/(?P<uuid>[^/]+)/mount/$',
            mount_repository, name='repository-mount'),
    re_path(r'^sys/backups/repositories/(?P<uuid>[^/]+)/umount/$',
            umount_repository, name='repository-umount'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for the Single Sign On app of FreedomBox.
"""

from captcha.fields import CaptchaField
from django import forms
from django.contrib.auth.forms import \
    AuthenticationForm as DjangoAuthenticationForm
from django.utils.translation import gettext_lazy as _


class AuthenticationForm(DjangoAuthenticationForm):
    """Authentication form with an additional username field attributes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['username'].widget.attrs.update({
            'autofocus': 'autofocus',
            'autocapitalize': 'none',
            'autocomplete': 'username'
        })


class CaptchaForm(forms.Form):
    """Form with a CAPTCHA field to use after 3 invalid login attempts."""
    captcha = CaptchaField(
        label=_('Enter the letters in the image to proceed to the login page'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for MediaWiki utility functions.
"""

import pathlib
from unittest.mock import patch

import pytest

from plinth.modules import mediawiki
from plinth.modules.mediawiki import privileged

pytestmark = pytest.mark.usefixtures('mock_privileged')
current_directory = pathlib.Path(__file__).parent
privileged_modules_to_mock = ['plinth.modules.mediawiki.privileged']


@pytest.fixture(name='test_configuration', autouse=True)
def fixture_test_configuration(tmp_path):
    """Use a separate MediaWiki configuration for tests.

    FreedomBoxStaticSettings.php is used read-only from source code location.
    """
    settings_file_name = 'FreedomBoxSettings.php'
    conf_file = tmp_path / settings_file_name
    conf_file.touch()
    with (patch('plinth.modules.mediawiki.USER_CONFIG_FILE', conf_file),
          patch('plinth.modules.mediawiki.privileged.CONF_FILE', conf_file)):
        yield


def test_private_mode():
    """Test enabling/disabling private mode."""
    assert not mediawiki.get_config()['enable_private_mode']
    privileged.set_private_mode(True)
    assert mediawiki.get_config()['enable_private_mode']
    privileged.set_private_mode(False)
    assert not mediawiki.get_config()['enable_private_mode']


def test_public_registrations():
    """Test enabling/disabling public registrations."""
    assert not mediawiki.get_config()['enable_public_registrations']
    privileged.set_public_registrations(True)
    assert mediawiki.get_config()['enable_public_registrations']
    privileged.set_public_registrations(False)
    assert not mediawiki.get_config()['enable_public_registrations']


def test_default_skin():
    """Test getting and setting the default skin."""
    assert mediawiki.get_config()['default_skin'] == 'timeless'
    new_skin = 'vector'
    privileged.set_default_skin(new_skin)
    assert mediawiki.get_config()['default_skin'] == new_skin


def test_domain():
    """Test getting and setting $wgServer."""
    assert mediawiki.get_config()['domain'] == 'freedombox.local'
    new_domain = 'mydomain.freedombox.rocks'
    mediawiki.set_server_url(new_domain)
    assert mediawiki.get_config()['domain'] == new_domain


def test_site_name():
    """Test getting and setting $wgSitename."""
    assert mediawiki.get_config()['site_name'] == 'Wiki'
    new_site_name = 'My MediaWiki'
    privileged.set_site_name(new_site_name)
    assert mediawiki.get_config()['site_name'] == new_site_name



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for wireguard.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.utils import format_lazy, import_from_gi

from . import manifest, utils

nm = import_from_gi('NM', '1.0')

_description = [
    _('WireGuard is a fast, modern, secure VPN tunnel.'),
    format_lazy(
        _('It can be used to connect to a VPN provider which supports '
          'WireGuard, and to route all outgoing traffic from {box_name} '
          'through the VPN.'), box_name=_(cfg.box_name)),
    format_lazy(
        _('A second use case is to connect a mobile device to {box_name} '
          'while travelling. While connected to a public Wi-Fi network, all '
          'traffic can be securely relayed through {box_name}.'),
        box_name=_(cfg.box_name))
]

SERVER_INTERFACE = 'wg0'


class WireguardApp(app_module.App):
    """FreedomBox app for wireguard."""

    app_id = 'wireguard'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('WireGuard'),
            icon_filename='wireguard',
            short_description=_('Virtual Private Network'),
            description=_description, manual_page='WireGuard',
            clients=manifest.clients,
            donation_url='https://www.wireguard.com/donations/')
        self.add(info)

        menu_item = menu.Menu('menu-wireguard', info.name,
                              info.short_description, info.icon_filename,
                              'wireguard:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-wireguard', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('wireguard:index'), login_required=True,
            clients=info.clients)
        self.add(shortcut)

        packages = Packages('packages-wireguard', ['wireguard'])
        self.add(packages)

        firewall = Firewall('firewall-wireguard', info.name,
                            ports=['wireguard-freedombox'], is_external=True)
        self.add(firewall)

    def enable(self):
        """Enable the app by simply storing a flag in key/value store."""
        from plinth import kvstore
        super().enable()
        kvstore.set('wireguard-enabled', True)
        utils.enable_connections(True)

    def disable(self):
        """Disable the app by simply storing a flag in key/value store."""
        from plinth import kvstore
        super().disable()
        kvstore.set('wireguard-enabled', False)
        utils.enable_connections(False)

    def is_enabled(self):
        """Return whether all leader components are enabled and flag is set."""
        from plinth import kvstore
        enabled = super().is_enabled()
        return enabled and kvstore.get_default('wireguard-enabled', False)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring Tor."""

import logging

from django.utils.translation import gettext_noop
from django.views.generic.edit import FormView

from plinth import app as app_module
from plinth import operation as operation_module
from plinth.modules import tor
from plinth.views import AppView

from . import privileged
from . import utils as tor_utils
from .forms import TorForm

config_process = None
logger = logging.getLogger(__name__)


class TorAppView(AppView):
    """Show Tor app main page."""

    app_id = 'tor'
    template_name = 'tor.html'
    form_class = TorForm
    prefix = 'tor'

    status = None

    def get_initial(self):
        """Return the values to fill in the form."""
        if not self.status:
            self.status = tor_utils.get_status()

        initial = super().get_initial()
        initial.update(self.status)
        return initial

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        if not self.status:
            self.status = tor_utils.get_status()

        context = super().get_context_data(*args, **kwargs)
        context['status'] = self.status
        return context

    def form_valid(self, form):
        """Configure tor app on successful form submission."""
        operation_module.manager.new('tor-configuration', self.app_id,
                                     gettext_noop('Updating configuration'),
                                     _apply_changes,
                                     [form.initial, form.cleaned_data],
                                     show_notification=False)
        # Skip check for 'Settings unchanged' message by calling grandparent
        return super(FormView, self).form_valid(form)


def _apply_changes(old_status, new_status):
    """Try to apply changes and handle errors."""
    logger.info('tor: applying configuration changes')
    exception_to_update = None
    message = None
    try:
        __apply_changes(old_status, new_status)
    except Exception as exception:
        exception_to_update = exception
        message = gettext_noop('Error configuring app: {error}').format(
            error=exception)
    else:
        message = gettext_noop('Configuration updated.')

    logger.info('tor: configuration changes completed')
    operation = operation_module.Operation.get_operation()
    operation.on_update(message, exception_to_update)


def __apply_changes(old_status, new_status):
    """Apply the changes."""
    arguments = {}
    app = app_module.App.get('tor')
    is_enabled = app.is_enabled()

    if old_status['relay_enabled'] != new_status['relay_enabled']:
        arguments['relay'] = new_status['relay_enabled']

    if old_status['bridge_relay_enabled'] != \
       new_status['bridge_relay_enabled']:
        arguments['bridge_relay'] = new_status['bridge_relay_enabled']

    if old_status['hs_enabled'] != new_status['hs_enabled']:
        arguments['hidden_service'] = new_status['hs_enabled']

    if old_status['use_upstream_bridges'] != \
       new_status['use_upstream_bridges']:
        arguments['use_upstream_bridges'] = new_status['use_upstream_bridges']

    if old_status['upstream_bridges'] != new_status['upstream_bridges']:
        arguments['upstream_bridges'] = new_status['upstream_bridges']

    if arguments:
        privileged.configure(**arguments)

        if is_enabled:
            privileged.restart()
            status = tor_utils.get_status()
            tor.update_hidden_service_domain(status)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for samba views.
"""

import pathlib
import urllib
from unittest.mock import patch

import pytest
from django import urls
from django.contrib.messages.storage.fallback import FallbackStorage

from plinth import module_loader
from plinth.modules.samba import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')
setfacl_path = pathlib.Path('/usr/bin/setfacl')

USERS = {"access_ok": ["testuser"], 'password_re_enter_needed': []}

DISKS = [{
    'device': '/dev/sda1',
    'label': '',
    'filesystem_type': 'ext4',
    'mount_point': '/',
    'percent_used': 63,
    'size_str': '9.5 GiB',
    'used_str': '5.7 GiB'
}, {
    'device': '/dev/sda2',
    'label': '',
    'filesystem_type': 'vfat',
    'mount_point': '/boot/efi',
    'percent_used': 50,
    'size_str': '150 MiB',
    'used_str': '75 MiB'
}]

SHARES = [
    {
        "name": "disk",
        "mount_point": "/",
        "path": "/var/lib/freedombox/shares/open_share",
        "share_type": "open"
    },
    {
        "name": "disk_home",
        "mount_point": "/",
        "path": "/var/lib/freedombox/shares/homes/%u",
        "share_type": "home"
    },
    {
        "name": "otherdisk",
        "mount_point": "/media/root/otherdisk",
        "path": "/media/root/otherdisk/FreedomBox/shares/homes/open_share",
        "share_type": "open"
    },
]


@pytest.fixture(autouse=True, scope='module')
def fixture_samba_urls():
    """Make sure samba app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.samba')
        module_loader.include_urls()
        yield


@pytest.fixture(autouse=True)
def samba_patch_privileged():
    """Patch privileged scripts runner."""
    with patch('plinth.modules.samba.privileged.get_shares') as get_shares, \
         patch('plinth.modules.samba.privileged.delete_share'), \
         patch('plinth.modules.samba.privileged.add_share'):
        get_shares.return_value = SHARES
        yield


def make_request(request, view, **kwargs):
    """Make request with a message storage."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)
    response = view(request, **kwargs)

    return response, messages


def test_samba_shares_view(rf):
    """Test that a share list has correct view data."""
    with (patch('plinth.views.AppView.get_context_data',
                return_value={'is_enabled': True}),
          patch('plinth.modules.samba.get_users', return_value=USERS),
          patch('plinth.modules.storage.get_mounts', return_value=DISKS),
          patch('plinth.views.AppView.app', return_value=None)):
        view = views.SambaAppView.as_view()
        response, _ = make_request(rf.get(''), view)

        assert response.context_data['disks'] == [DISKS[0]]
        assert response.context_data['shared_mounts'] == {
            '/': ['open', 'home'],
            '/media/root/otherdisk': ['open']
        }
        assert response.context_data['unavailable_shares'] == [{
            'mount_point': '/media/root/otherdisk',
            'name': 'otherdisk',
            'path': '/media/root/otherdisk/FreedomBox/shares/homes/open_share',
            'share_type': 'open'
        }]
        assert response.context_data['users'] == USERS
        assert response.status_code == 200


@pytest.mark.skipif(not setfacl_path.exists(), reason='setfacl not installed')
def test_enable_samba_share_view(rf):
    """Test that enabling share sends correct success message."""
    form_data = {'filesystem_type': 'ext4', 'open_share': 'enable'}
    mount_point = urllib.parse.quote('/')
    response, messages = make_request(rf.post('', data=form_data), views.share,
                                      mount_point=mount_point)

    assert list(messages)[0].message == 'Share enabled.'
    assert response.status_code == 302
    assert response.url == urls.reverse('samba:index')


def test_enable_samba_share_failed_view(rf):
    """Test that share enabling failure sends correct error message."""
    form_data = {'filesystem_type': 'ext4', 'open_share': 'enable'}
    mount_point = urllib.parse.quote('/')
    error_message = 'Sharing failed'
    with patch('plinth.modules.samba.add_share',
               side_effect=RuntimeError(error_message)):
        response, messages = make_request(rf.post('', data=form_data),
                                          views.share, mount_point=mount_point)

        assert list(messages)[0].message == 'Error enabling share: {0}'.format(
            error_message)
        assert response.status_code == 302
        assert response.url == urls.reverse('samba:index')


def test_disable_samba_share(rf):
    """Test that enabling share sends correct success message."""
    form_data = {'filesystem_type': 'ext4', 'open_share': 'disable'}
    mount_point = urllib.parse.quote('/')
    response, messages = make_request(rf.post('', data=form_data), views.share,
                                      mount_point=mount_point)

    assert list(messages)[0].message == 'Share disabled.'
    assert response.status_code == 302
    assert response.url == urls.reverse('samba:index')


def test_disable_samba_share_failed_view(rf):
    """Test that share disabling failure sends correct error message."""
    form_data = {'filesystem_type': 'ext4', 'open_share': 'disable'}
    mount_point = urllib.parse.quote('/')
    error_message = 'Unsharing failed'
    with patch('plinth.modules.samba.privileged.delete_share',
               side_effect=RuntimeError(error_message)):
        response, messages = make_request(rf.post('', data=form_data),
                                          views.share, mount_point=mount_point)

        assert list(messages)[
            0].message == 'Error disabling share: {0}'.format(error_message)
        assert response.status_code == 302
        assert response.url == urls.reverse('samba:index')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Dummy file to help pytest-django path detection.

pytest-django searches for a folder with manage.py and treats that as parent
directory for Django project. This folder is then added to Python path managed
in sys.path. This allows the Django setting module to be discovered as
plinth.tests.data.django_test_settings. pytest can then be invoked simply as
'py.test-3' instead of 'python3 -m pytest'.
"""



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test the Let's Encrypt component for managing certificates.
"""

from unittest.mock import call, patch

import pytest

from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.names.components import DomainName, DomainType


@pytest.fixture(name='empty_letsencrypt_list', autouse=True)
def fixture_empty_letsencrypt_list():
    """Remove all entries from Let's Encrypt component list."""
    LetsEncrypt._all = {}


@pytest.fixture(name='component')
def fixture_component():
    """Create a new component for testing."""
    return LetsEncrypt(
        'test-component', domains=['valid.example', 'invalid.example'],
        daemons=['test-daemon'], should_copy_certificates=True,
        private_key_path='/etc/test-app/{domain}/private.path',
        certificate_path='/etc/test-app/{domain}/certificate.path',
        user_owner='test-user', group_owner='test-group',
        managing_app='test-app')


@pytest.fixture(name='try_restart')
def fixture_try_restart():
    """Patch and return service.try_restart privileged call."""
    with patch('plinth.privileged.service.try_restart') as try_restart:
        yield try_restart


@pytest.fixture(name='copy_certificate')
def fixture_copy_certificate():
    """Patch and return privileged.copy_certificate call."""
    with patch('plinth.modules.letsencrypt.privileged.copy_certificate'
               ) as copy_certificate:
        yield copy_certificate


@pytest.fixture(name='compare_certificate')
def fixture_compare_certificate():
    """Patch and return privileged.compare_certificate call."""
    with patch('plinth.modules.letsencrypt.privileged.compare_certificate'
               ) as compare_certificate:
        yield compare_certificate


@pytest.fixture(name='get_status')
def fixture_get_status():
    """Return patched letsencrypt.get_status() method."""
    domains = ['valid.example']
    with patch('plinth.modules.letsencrypt.get_status') as get_status:
        get_status.return_value = {
            'domains': {
                domain: {
                    'lineage': '/etc/letsencrypt/live/' + domain,
                    'validity': 'valid',
                }
                for domain in domains
            }
        }
        yield get_status


@pytest.fixture(name='domain_list')
def fixture_domain_list():
    """Return patch DomainName.list() method."""
    method = 'plinth.modules.names.components.DomainName.list'
    with patch(method) as domain_list:
        DomainType._all = {}
        DomainType('domain-type-1', 'type-1', 'url1', False)
        DomainType('domain-type-2', 'type-2', 'url1', True)
        domain1 = DomainName('domain-name-1', 'invalid1.example',
                             'domain-type-1', '__all__')
        domain2 = DomainName('domain-name-2', 'valid.example', 'domain-type-2',
                             '__all__')
        domain3 = DomainName('domain-name-3', 'invalid2.example',
                             'domain-type-2', '__all__')
        domain_list.return_value = [domain1, domain2, domain3]
        yield domain_list


def test_init_without_arguments():
    """Test that component is initialized with defaults properly."""
    component = LetsEncrypt('test-component')

    assert component.component_id == 'test-component'
    assert component.domains is None
    assert component.daemons is None
    assert not component.should_copy_certificates
    assert component.private_key_path is None
    assert component.certificate_path is None
    assert component.user_owner is None
    assert component.group_owner is None
    assert component.managing_app is None
    assert len(component._all) == 1
    assert component._all['test-component'] == component


def test_init(component):
    """Test initializing the component."""
    assert component.domains == ['valid.example', 'invalid.example']
    assert component.daemons == ['test-daemon']
    assert component.should_copy_certificates
    assert component.private_key_path == '/etc/test-app/{domain}/private.path'
    assert component.certificate_path == \
        '/etc/test-app/{domain}/certificate.path'
    assert component.user_owner == 'test-user'
    assert component.group_owner == 'test-group'
    assert component.managing_app == 'test-app'


def test_init_values():
    """Test initializing with invalid values."""
    properties = {
        'private_key_path': 'test-private-key-path',
        'certificate_path': 'test-certificate-path',
        'user_owner': 'test-user',
        'group_owner': 'test-group',
        'managing_app': 'test-app'
    }
    LetsEncrypt('test-component', should_copy_certificates=True, **properties)
    for key in properties:
        new_properties = dict(properties)
        new_properties[key] = None
        with pytest.raises(ValueError):
            LetsEncrypt('test-component', should_copy_certificates=True,
                        **new_properties)


def test_domains():
    """Test getting domains."""
    component = LetsEncrypt('test-component', domains=lambda: ['test-domains'])
    assert component.domains == ['test-domains']


def test_list():
    """Test listing components."""
    component1 = LetsEncrypt('test-component1')
    component2 = LetsEncrypt('test-component2')
    assert set(LetsEncrypt.list()) == {component1, component2}


def _assert_copy_certificate_called(component, copy_certificate, domains):
    """Check that copy certificate calls have been made properly."""
    expected_calls = []
    for domain, domain_status in domains.items():
        if domain_status == 'valid':
            source_private_key_path = \
                '/etc/letsencrypt/live/{}/privkey.pem'.format(domain)
            source_certificate_path = \
                '/etc/letsencrypt/live/{}/fullchain.pem'.format(domain)
        else:
            source_private_key_path = '/etc/ssl/private/ssl-cert-snakeoil.key'
            source_certificate_path = '/etc/ssl/certs/ssl-cert-snakeoil.pem'

        private_key_path = '/etc/test-app/{}/private.path'.format(domain)
        certificate_path = '/etc/test-app/{}/certificate.path'.format(domain)
        expected_call = call(component.managing_app,
                             str(source_private_key_path),
                             str(source_certificate_path), private_key_path,
                             certificate_path, component.user_owner,
                             component.group_owner)
        expected_calls.append(expected_call)

    copy_certificate.assert_has_calls(expected_calls, any_order=True)


def _assert_restarted_daemons(daemons, try_restart):
    """Check that a call has restarted the daemons of a component."""
    expected_calls = [call(daemon) for daemon in daemons]
    try_restart.assert_has_calls(expected_calls, any_order=True)


def test_setup_certificates(copy_certificate, try_restart, get_status,
                            component):
    """Test that initial copying of certs for an app works."""
    component.setup_certificates()
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
        'invalid.example': 'invalid'
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_setup_certificates_without_copy(copy_certificate, try_restart,
                                         get_status, component):
    """Test that initial copying of certs for an app works."""
    component.should_copy_certificates = False
    component.setup_certificates()
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons(component.daemons, try_restart)


def test_setup_certificates_with_app_domains(copy_certificate, try_restart,
                                             get_status, component):
    """Test that initial copying of certs for an app works."""
    component._domains = ['irrelevant1.example', 'irrelevant2.example']
    component.setup_certificates(
        app_domains=['valid.example', 'invalid.example'])
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
        'invalid.example': 'invalid'
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_setup_certificates_with_all_domains(domain_list, copy_certificate,
                                             try_restart, get_status,
                                             component):
    """Test that initial copying for certs works when app domains is '*'."""
    component._domains = '*'
    component.setup_certificates()
    _assert_copy_certificate_called(
        component, copy_certificate, {
            'valid.example': 'valid',
            'invalid1.example': 'invalid',
            'invalid2.example': 'invalid'
        })
    _assert_restarted_daemons(component.daemons, try_restart)


def _assert_compare_certificate_called(component, compare_certificate,
                                       domains):
    """Check that compare certificate was called properly."""
    expected_calls = []
    for domain in domains:
        source_private_key_path = \
            '/etc/letsencrypt/live/{}/privkey.pem'.format(domain)
        source_certificate_path = \
            '/etc/letsencrypt/live/{}/fullchain.pem'.format(domain)
        private_key_path = '/etc/test-app/{}/private.path'.format(domain)
        certificate_path = '/etc/test-app/{}/certificate.path'.format(domain)
        expected_call = call(component.managing_app,
                             str(source_private_key_path),
                             str(source_certificate_path), private_key_path,
                             certificate_path)
        expected_calls.append(expected_call)

    compare_certificate.assert_has_calls(expected_calls, any_order=True)


def test_get_status(component, compare_certificate, get_status):
    """Test that getting domain status works."""
    compare_certificate.return_value = True
    assert component.get_status() == {
        'valid.example': 'valid',
        'invalid.example': 'self-signed'
    }
    _assert_compare_certificate_called(component, compare_certificate,
                                       ['valid.example'])


def test_get_status_outdate_copy(component, compare_certificate, get_status):
    """Test that getting domain status works with outdated copy."""
    compare_certificate.return_value = False
    assert component.get_status() == {
        'valid.example': 'outdated-copy',
        'invalid.example': 'self-signed'
    }
    _assert_compare_certificate_called(component, compare_certificate,
                                       ['valid.example'])


def test_get_status_without_copy(component, get_status):
    """Test that getting domain status works without copying."""
    component.should_copy_certificates = False
    assert component.get_status() == {
        'valid.example': 'valid',
        'invalid.example': 'self-signed'
    }


def test_on_certificate_obtained(copy_certificate, try_restart, component):
    """Test that certificate obtained event handler works."""
    component.on_certificate_obtained(['valid.example', 'irrelevant.example'],
                                      '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_obtained_with_all_domains(copy_certificate,
                                                  try_restart, component):
    """Test that certificate obtained event handler works for app with
       all domains.
    """
    component._domains = '*'
    component.on_certificate_obtained(['valid.example'],
                                      '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_obtained_irrelevant(copy_certificate, try_restart,
                                            component):
    """Test that certificate obtained event handler works with
       irrelevant domain.
    """
    component.on_certificate_obtained(
        ['irrelevant.example'], '/etc/letsencrypt/live/irrelevant.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons([], try_restart)


def test_on_certificate_obtained_without_copy(copy_certificate, try_restart,
                                              component):
    """Test that certificate obtained event handler works without copying."""
    component.should_copy_certificates = False
    component.on_certificate_obtained(['valid.example'],
                                      '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_renewed(copy_certificate, try_restart, component):
    """Test that certificate renewed event handler works."""
    component.on_certificate_renewed(['valid.example', 'irrelevant.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_renewed_irrelevant(copy_certificate, try_restart,
                                           component):
    """Test that certificate renewed event handler works for
       irrelevant domains.
"""
    component.on_certificate_renewed(
        ['irrelevant.example'], '/etc/letsencrypt/live/irrelevant.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons([], try_restart)


def test_on_certificate_renewed_without_copy(copy_certificate, try_restart,
                                             component):
    """Test that certificate renewed event handler works without copying."""
    component.should_copy_certificates = False
    component.on_certificate_renewed(['valid.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_revoked(copy_certificate, try_restart, component):
    """Test that certificate revoked event handler works."""
    component.on_certificate_revoked(['valid.example', 'irrelevant.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'invalid',
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_revoked_irrelevant(copy_certificate, try_restart,
                                           component):
    """Test that certificate revoked event handler works for
       irrelevant domains.
    """
    component.on_certificate_revoked(
        ['irrelevant.example'], '/etc/letsencrypt/live/irrelevant.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons([], try_restart)


def test_on_certificate_revoked_without_copy(copy_certificate, try_restart,
                                             component):
    """Test that certificate revoked event handler works without copying."""
    component.should_copy_certificates = False
    component.on_certificate_revoked(['valid.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_deleted(copy_certificate, try_restart, component):
    """Test that certificate deleted event handler works."""
    component.on_certificate_deleted(['valid.example', 'irrelevant.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'invalid',
    })
    _assert_restarted_daemons(component.daemons, try_restart)


def test_on_certificate_deleted_irrelevant(copy_certificate, try_restart,
                                           component):
    """Test that certificate deleted event handler works for
       irrelevant domains.
    """
    component.on_certificate_deleted(
        ['irrelevant.example'], '/etc/letsencrypt/live/irrelevant.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons([], try_restart)


def test_on_certificate_deleted_without_copy(copy_certificate, try_restart,
                                             component):
    """Test that certificate deleted event handler works without copying."""
    component.should_copy_certificates = False
    component.on_certificate_deleted(['valid.example'],
                                     '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})
    _assert_restarted_daemons(component.daemons, try_restart)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Let's Encrypt module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/letsencrypt/$', views.LetsEncryptAppView.as_view(),
            name='index'),
    re_path(r'^sys/letsencrypt/obtain/(?P<domain>[^/]+)/$', views.obtain,
            name='obtain'),
    re_path(r'^sys/letsencrypt/re-obtain/(?P<domain>[^/]+)/$', views.reobtain,
            name='re-obtain'),
    re_path(r'^sys/letsencrypt/revoke/(?P<domain>[^/]+)/$', views.revoke,
            name='revoke'),
    re_path(r'^sys/letsencrypt/delete/(?P<domain>[^/]+)/$', views.delete,
            name='delete'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for samba module."""

import logging
import os
import urllib.parse
from collections import defaultdict

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST

from plinth import views
from plinth.modules import samba, storage

from . import privileged

logger = logging.getLogger(__name__)


def get_share_mounts():
    """Return list of shareable mounts."""
    ignore_mounts = ('/boot', '/boot/efi', '/boot/firmware', '/.snapshots')
    mounts = []

    for mount in storage.get_mounts():
        mount_point = mount['mount_point']
        if mount_point in ignore_mounts or os.path.isfile(mount_point):
            continue
        basename = os.path.basename(mount_point)
        mount['name'] = basename or _('FreedomBox OS disk')
        mount['share_name_prefix'] = basename or 'disk'
        mounts.append(mount)

    return sorted(mounts, key=lambda k: k['mount_point'])


class SambaAppView(views.AppView):
    """Samba sharing basic configuration."""

    app_id = 'samba'
    template_name = 'samba.html'

    def get_context_data(self, *args, **kwargs):
        """Return template context data."""
        context = super().get_context_data(*args, **kwargs)
        disks = get_share_mounts()
        context['disks'] = disks

        shares = privileged.get_shares()
        shared_mounts = defaultdict(list)
        for share in shares:
            shared_mounts[share['mount_point']].append(share['share_type'])
        context['shared_mounts'] = shared_mounts

        context['share_types'] = [{
            'id': 'open',
            'type': _('Open Share'),
            'share_name_suffix': ''
        }, {
            'id': 'group',
            'type': _('Group Share'),
            'share_name_suffix': '_group'
        }, {
            'id': 'home',
            'type': _('Home Share'),
            'share_name_suffix': '_home'
        }]

        unavailable_shares = []
        for share in shares:
            for disk in disks:
                if share['mount_point'] == disk['mount_point']:
                    break
            else:
                unavailable_shares.append(share)
        context['unavailable_shares'] = unavailable_shares

        context['users'] = samba.get_users()

        return context


@require_POST
def share(request, mount_point):
    """Enable sharing, given its root path.

    mount_point is urlquoted.

    """
    mount_point = urllib.parse.unquote(mount_point)
    filesystem = request.POST.get('filesystem_type', '')

    share_types = ['open', 'group', 'home']

    for share_type in share_types:
        action = request.POST.get(share_type + '_share', '')
        if action == 'enable':
            try:
                samba.add_share(mount_point, share_type, filesystem)
                messages.success(request, _('Share enabled.'))
            except Exception as exception:
                logger.exception('Error enabling share')
                messages.error(
                    request,
                    _('Error enabling share: {error_message}').format(
                        error_message=exception))
        elif action == 'disable':
            try:
                privileged.delete_share(mount_point, share_type)
                messages.success(request, _('Share disabled.'))
            except Exception as exception:
                logger.exception('Error disabling share')
                messages.error(
                    request,
                    _('Error disabling share: {error_message}').format(
                        error_message=exception))

    return redirect(reverse('samba:index'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for email app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.email]


class TestEmailApp(functional.BaseAppTests):
    app_name = 'email'
    has_service = True
    has_web = False
    check_diagnostics = True

    @pytest.fixture(autouse=True)
    def fixture_background(self, session_browser):
        functional.login(session_browser)
        functional.install(session_browser, self.app_name)
        functional.app_enable(session_browser, self.app_name)
        yield
        functional.login(session_browser)
        functional.app_disable(session_browser, self.app_name)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure I2P."""

from plinth.actions import privileged
from plinth.modules.i2p.helpers import RouterEditor, TunnelEditor


@privileged
def set_tunnel_property(name: str, property_: str, value: str):
    """Modify the configuration file for a certain tunnel."""
    editor = TunnelEditor()
    editor \
        .read_conf() \
        .set_tunnel_idx(name) \
        .set_tunnel_prop(property_, value) \
        .write_conf()


@privileged
def add_favorite(name: str, url: str, description: str | None,
                 icon: str | None):
    """Add a favorite to router.config."""
    editor = RouterEditor()
    editor.read_conf().add_favorite(name, url, description, icon).write_conf()



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('Searx'),
    'platforms': [{
        'type': 'web',
        'url': '/searx/'
    }]
}]

PUBLIC_ACCESS_SETTING_FILE = '/etc/searx/allow_public_access'

backup = {'config': {'files': [PUBLIC_ACCESS_SETTING_FILE]}}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for i2p app.
"""

import pytest

from plinth.tests.functional import BaseAppTests

pytestmark = [pytest.mark.apps, pytest.mark.i2p]


class TestI2pApp(BaseAppTests):
    app_name = 'i2p'
    has_service = True
    has_web = True
    diagnostics_delay = 1



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for I2P.
"""

from django.utils.translation import gettext_lazy as _

_package_id = 'net.geti2p.i2p'
_download_url = 'https://geti2p.net/download'

clients = [{
    'name':
        _('I2P'),
    'platforms': [{
        'type': 'web',
        'url': '/i2p/'
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'i2p',
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _download_url,
    }, {
        'type': 'download',
        'os': 'macos',
        'url': _download_url,
    }, {
        'type': 'download',
        'os': 'windows',
        'url': _download_url,
    }]
}]

backup = {
    'secrets': {
        'directories': ['/var/lib/i2p/i2p-config']
    },
    'services': ['i2p']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure privoxy."""

import pathlib

import augeas

from plinth import action_utils
from plinth.actions import privileged

PRIVOXY_CONF_PATH = pathlib.Path('/etc/privoxy/config')


@privileged
def pre_install():
    """Preseed debconf values before packages are installed."""
    action_utils.debconf_set_selections(
        ['privoxy privoxy/listen-address string [::]:8118'])


@privileged
def setup():
    """Setup Privoxy configuration after installing it."""
    _restrict_access()


def _load_augeus():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Spacevars', str(PRIVOXY_CONF_PATH))
    aug.set('/augeas/context', '/files' + str(PRIVOXY_CONF_PATH))
    aug.load()
    return aug


def _restrict_access():
    """Make sure Privoxy isn't available over the Internet."""
    # https://en.wikipedia.org/wiki/localhost
    # https://en.wikipedia.org/wiki/Private_network
    # https://en.wikipedia.org/wiki/Link-local_address
    # https://en.wikipedia.org/wiki/Unique_local_address
    ip_ranges = [
        '127.0.0.0/8',  # IPv4 loopback address
        '10.0.0.0/8',  # IPv4 private address
        '172.16.0.0/12',  # IPv4 private address
        '192.168.0.0/16',  # IPv4 private address
        '169.254.0.0/16',  # IPv4 auto-configuration
        '[::1]',  # IPv4 loopback address
        '[fc00::]/7',  # IPv6 unique local addresses
        '[fe80::]/10',  # IPv6 auto-configuration
    ]
    aug = _load_augeus()
    for ip_range in ip_ranges:
        matches = [
            match for match in aug.match('permit-access')
            if aug.get(match) == ip_range
        ]
        if not any(matches):
            aug.set('permit-access[last() + 1]', ip_range)

    aug.save()



# SPDX-License-Identifier: AGPL-3.0-or-later

#
# Generated by Django 2.2.6 on 2020-01-29 03:17
#
"""
Django migration for adding the notification model.
"""

from django.db import migrations, models

from plinth.models import JSONField


class Migration(migrations.Migration):

    dependencies = [
        ('plinth', '0004_userprofile'),
    ]

    operations = [
        migrations.CreateModel(
            name='StoredNotification',
            fields=[
                ('id',
                 models.CharField(max_length=128, primary_key=True,
                                  serialize=False)),
                ('app_id',
                 models.CharField(default=None, max_length=128, null=True)),
                ('severity', models.CharField(max_length=32)),
                ('title', models.CharField(max_length=256)),
                ('message', models.TextField(default=None, null=True)),
                ('actions', JSONField(default=list)),
                ('body_template',
                 models.CharField(max_length=128, null=True, default=None)),
                ('data', JSONField(default=dict)),
                ('created_time', models.DateTimeField(auto_now_add=True)),
                ('last_update_time', models.DateTimeField(auto_now=True)),
                ('user',
                 models.CharField(max_length=128, default=None, null=True)),
                ('group',
                 models.CharField(max_length=128, default=None, null=True)),
                ('dismissed', models.BooleanField(default=False)),
            ],
        ),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox views for basic system configuration."""

import logging

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views
from plinth.modules import config
from plinth.signals import (domain_added, domain_removed, post_hostname_change,
                            pre_hostname_change)

from . import privileged
from .forms import ConfigurationForm

LOGGER = logging.getLogger(__name__)


class ConfigAppView(views.AppView):
    """Serve configuration page."""

    form_class = ConfigurationForm
    app_id = 'config'

    def get_initial(self):
        """Return the current status."""
        return {
            'hostname': config.get_hostname(),
            'domainname': config.get_domainname(),
            'homepage': config.get_home_page(),
            'advanced_mode': config.get_advanced_mode(),
            'logging_mode': privileged.get_logging_mode(),
        }

    def form_valid(self, form):
        """Apply the form changes."""
        old_status = form.initial
        new_status = form.cleaned_data

        is_changed = False

        if old_status['hostname'] != new_status['hostname']:
            try:
                set_hostname(new_status['hostname'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error setting hostname: {exception}').format(
                        exception=exception))
            else:
                messages.success(self.request, _('Hostname set'))

        if old_status['domainname'] != new_status['domainname']:
            try:
                set_domainname(new_status['domainname'],
                               old_status['domainname'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error setting domain name: {exception}').format(
                        exception=exception))
            else:
                messages.success(self.request, _('Domain name set'))

        if old_status['homepage'] != new_status['homepage']:
            try:
                config.change_home_page(new_status['homepage'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error setting webserver home page: {exception}').format(
                        exception=exception))
            else:
                messages.success(self.request, _('Webserver home page set'))

        if old_status['advanced_mode'] != new_status['advanced_mode']:
            try:
                config.set_advanced_mode(new_status['advanced_mode'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error changing advanced mode: {exception}').format(
                        exception=exception))
            else:
                if new_status['advanced_mode']:
                    messages.success(self.request,
                                     _('Showing advanced apps and features'))
                else:
                    messages.success(self.request,
                                     _('Hiding advanced apps and features'))

        if old_status['logging_mode'] != new_status['logging_mode']:
            privileged.set_logging_mode(new_status['logging_mode'])
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)


def set_hostname(hostname):
    """Set machine hostname and send signals before and after."""
    old_hostname = config.get_hostname()
    domainname = config.get_domainname()

    # Hostname should be ASCII. If it's unicode but passed our
    # valid_hostname check, convert
    hostname = str(hostname)

    pre_hostname_change.send_robust(sender='config', old_hostname=old_hostname,
                                    new_hostname=hostname)

    LOGGER.info('Changing hostname to - %s', hostname)
    privileged.set_hostname(hostname)

    LOGGER.info('Setting domain name after hostname change - %s', domainname)
    privileged.set_domainname(domainname)

    post_hostname_change.send_robust(sender='config',
                                     old_hostname=old_hostname,
                                     new_hostname=hostname)


def set_domainname(domainname, old_domainname):
    """Set machine domain name to domainname."""
    old_domainname = config.get_domainname()

    # Domain name is not case sensitive, but Let's Encrypt certificate
    # paths use lower-case domain name.
    domainname = domainname.lower()

    LOGGER.info('Changing domain name to - %s', domainname)
    privileged.set_domainname(domainname)

    # Update domain registered with Name Services module.
    if old_domainname:
        domain_removed.send_robust(sender='config',
                                   domain_type='domain-type-static',
                                   name=old_domainname)

    if domainname:
        domain_added.send_robust(sender='config',
                                 domain_type='domain-type-static',
                                 name=domainname, services='__all__')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the Ejabberd app."""

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth.modules import coturn, ejabberd
from plinth.modules.coturn.components import TurnConfiguration
from plinth.views import AppView

from . import privileged
from .forms import EjabberdForm


class EjabberdAppView(AppView):
    """Show ejabberd as a service."""

    app_id = 'ejabberd'
    template_name = 'ejabberd.html'
    form_class = EjabberdForm

    def get_initial(self):
        """Return initial data to fill in the form."""
        config, managed = ejabberd.get_turn_configuration()
        return super().get_initial() | {
            'domain_names': ejabberd.get_domains(),
            'MAM_enabled': privileged.mam('status'),
            'enable_managed_turn': managed,
            'turn_uris': '\n'.join(config.uris),
            'shared_secret': config.shared_secret
        }

    def get_context_data(self, *args, **kwargs):
        """Add service to the context data."""
        context = super().get_context_data(*args, **kwargs)
        domains = ejabberd.get_domains()
        context['domainname'] = domains[0] if domains else None
        return context

    @staticmethod
    def _handle_domain_names_configuration(new_config):
        """Update list of domain names in configuration."""
        ejabberd.set_domains(new_config['domain_names'])

    @staticmethod
    def _handle_turn_configuration(old_config, new_config):
        if not new_config['enable_managed_turn']:
            new_turn_uris = new_config['turn_uris'].splitlines()
            new_shared_secret = new_config['shared_secret']

            turn_config_changed = \
                old_config['turn_uris'] != new_turn_uris or \
                old_config['shared_secret'] != new_shared_secret

            if turn_config_changed:
                ejabberd.update_turn_configuration(
                    TurnConfiguration(None, new_turn_uris, new_shared_secret),
                    managed=False)
        else:
            ejabberd.update_turn_configuration(coturn.get_config(),
                                               managed=True)

    @staticmethod
    def _handle_MAM_configuration(old_config, new_config):
        # note ejabberd action "enable" or "disable" restarts, if running
        if new_config['MAM_enabled']:
            privileged.mam('enable')
        else:
            privileged.mam('disable')

    def form_valid(self, form):
        """Enable/disable a service and set messages."""
        old_config = form.initial
        new_config = form.cleaned_data

        def changed(prop):
            return old_config[prop] != new_config[prop]

        is_changed = False

        if changed('domain_names'):
            self._handle_domain_names_configuration(new_config)
            is_changed = True

        if changed('MAM_enabled'):
            self._handle_MAM_configuration(old_config, new_config)
            is_changed = True

        if changed('enable_managed_turn') or changed('turn_uris') or \
           changed('shared_secret'):
            self._handle_turn_configuration(old_config, new_config)
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for config app.
"""

import pytest

from plinth.tests import functional

pytestmark = [
    pytest.mark.system, pytest.mark.essential, pytest.mark.domain,
    pytest.mark.config
]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login."""
    functional.login(session_browser)


def test_change_hostname(session_browser):
    """Test changing the hostname."""
    functional.set_hostname(session_browser, 'mybox')
    assert _get_hostname(session_browser) == 'mybox'


def test_change_domain_name(session_browser):
    """Test changing the domain name."""
    functional.set_domain_name(session_browser, 'mydomain.example')
    assert _get_domain_name(session_browser) == 'mydomain.example'

    # Capitalization is ignored.
    functional.set_domain_name(session_browser, 'Mydomain.example')
    assert _get_domain_name(session_browser) == 'mydomain.example'


def test_change_home_page(session_browser):
    """Test changing webserver home page."""
    functional.install(session_browser, 'syncthing')
    functional.app_enable(session_browser, 'syncthing')
    _set_home_page(session_browser, 'syncthing')

    _set_home_page(session_browser, 'plinth')
    assert _check_home_page_redirect(session_browser, 'plinth')


def _get_hostname(browser):
    functional.nav_to_module(browser, 'config')
    return browser.find_by_id('id_hostname').value


def _get_domain_name(browser):
    functional.nav_to_module(browser, 'config')
    return browser.find_by_id('id_domainname').value


def _set_home_page(browser, home_page):
    if 'plinth' not in home_page and 'apache' not in home_page:
        home_page = 'shortcut-' + home_page

    functional.nav_to_module(browser, 'config')
    drop_down = browser.find_by_id('id_homepage')
    drop_down.select(home_page)
    functional.submit(browser, form_class='form-configuration')


def _check_home_page_redirect(browser, app_name):
    functional.visit(browser, '/')
    return browser.find_by_xpath(
        "//a[contains(@href, '/plinth/') and @title='FreedomBox']")



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
GnuDIP client for updating Dynamic DNS records.
"""

import hashlib
import logging
import socket as socket_module

BUF_SIZE = 128
GNUDIP_PORT = 3495
TIMEOUT = 10

logger = logging.getLogger(__name__)


def update(server, domain, username, password):
    """Update Dynamic DNS record."""
    domain = domain.removeprefix(username + '.')
    password_digest = hashlib.md5(password.encode()).hexdigest()

    with socket_module.socket(socket_module.AF_INET,
                              socket_module.SOCK_STREAM) as socket:
        logger.debug('Connecting to %s:%d, timeout %ss', server, GNUDIP_PORT,
                     TIMEOUT)
        socket.settimeout(TIMEOUT)
        socket.connect((server, GNUDIP_PORT))
        salt = socket.recv(BUF_SIZE).decode().strip()
        salted_digest = password_digest + '.' + salt
        final_digest = hashlib.md5(salted_digest.encode()).hexdigest()

        update_request = username + ':' + final_digest + ':' + domain + ':2\n'
        socket.sendall(update_request.encode())

        response = socket.recv(BUF_SIZE).decode().strip()
        result, _, new_ip = response.partition(':')
        result = (int(result) == 0)

        new_ip = new_ip if result else None
        return result, new_ip



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure minidlna server."""

import subprocess
from os import chmod, fdopen, remove, stat
from shutil import move
from tempfile import mkstemp

import augeas

from plinth import action_utils
from plinth.actions import privileged
from plinth.utils import grep

CONFIG_PATH = '/etc/minidlna.conf'

SYSCTL_CONF = '''# This file is managed and overwritten by FreedomBox.
# Helps minidlna monitor changes in large media directories
fs.inotify.max_user_watches = 100000
'''


def _undo_old_configuration_changes():
    """Restore /etc/sysctl.conf to before our changes.

    Older version of minidlna app in FreedomBox < 20.9 wrote to
    /etc/sysctl.conf directly. This will cause conffile prompt during upgrade
    of procps package. Undo the changes so that upgrade can happen smoothly.

    """
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Sysctl/lens', 'Sysctl.lns')
    aug.set('/augeas/load/Sysctl/incl[last() + 1]', '/etc/sysctl.conf')
    aug.load()

    key_path = '/files/etc/sysctl.conf/fs.inotify.max_user_watches'
    if aug.get(key_path) == '100000':
        aug.remove(key_path)
        aug.save()


@privileged
def setup():
    """Increase inotify watches per folder.

    This is to allow minidlna to monitor changes in large media-dirs.
    """
    _undo_old_configuration_changes()
    with open('/etc/sysctl.d/50-freedombox-minidlna.conf', 'w',
              encoding='utf-8') as conf:
        conf.write(SYSCTL_CONF)

    subprocess.run(['systemctl', 'restart', 'systemd-sysctl'], check=True)


@privileged
def get_media_dir() -> str:
    """Retrieve media directory from minidlna.conf."""
    line = grep('^media_dir=', CONFIG_PATH)
    return line[0].split('=')[1]


@privileged
def set_media_dir(media_dir: str):
    """Set media directory in minidlna.conf."""
    line = grep('^media_dir=', CONFIG_PATH)[0]

    new_line = 'media_dir=%s\n' % media_dir
    replace_in_config_file(CONFIG_PATH, line, new_line)
    if action_utils.service_is_running('minidlna'):
        action_utils.service_restart('minidlna')


def replace_in_config_file(file_path, pattern, subst):
    """Replace a directive in configuration file.

    - Create a temporary minidlna.conf file
    - Replace the media dir config
    - Remove original one and move the temporary file
    - Preserve permissions as the original file
    """
    temp_file, temp_file_path = mkstemp()
    with fdopen(temp_file, 'w') as new_file:
        with open(file_path, encoding='utf-8') as old_file:
            for line in old_file:
                new_file.write(line.replace(pattern, subst))

    old_st_mode = stat(file_path).st_mode
    remove(file_path)
    move(temp_file_path, file_path)
    chmod(file_path, old_st_mode)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Kiwix module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/kiwix/$', views.KiwixAppView.as_view(), name='index'),
    re_path(r'^apps/kiwix/package/add/$', views.AddPackageView.as_view(),
            name='add-package'),
    re_path(r'^apps/kiwix/package/(?P<zim_id>[a-zA-Z0-9-]+)/delete/$',
            views.delete_package, name='delete-package'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for custom middleware.
"""

from unittest.mock import MagicMock, Mock, call, patch

import pytest
from django.contrib.auth.models import AnonymousUser, Group, User
from django.core.exceptions import PermissionDenied
from django.db.utils import OperationalError
from django.http import HttpResponse
from django.test.client import RequestFactory
from stronghold.decorators import public

from plinth import app as app_module
from plinth.middleware import (AdminRequiredMiddleware, CommonErrorMiddleware,
                               SetupMiddleware)

setup_helper = None


@pytest.fixture(name='kwargs')
def fixture_kwargs():
    """Fixture for returning kwargs for creating middleware."""
    return {
        'view_func': HttpResponse,
        'view_args': [],
        'view_kwargs': {},
    }


@pytest.fixture(name='app')
def fixture_app():
    """Fixture for returning a test app."""

    class AppTest(app_module.App):
        """Test app."""
        app_id = 'mockapp'

        def __init__(self):
            """Add info component."""
            super().__init__()
            self.add(app_module.Info('mockapp', version=1))

        def get_setup_state(self):
            return app_module.App.SetupState.NEEDS_SETUP

    app_module.App._all_apps = {}
    return AppTest()


class TestSetupMiddleware:
    """Test cases for setup middleware."""

    @staticmethod
    @pytest.fixture(name='middleware')
    def fixture_middleware(load_cfg):
        """Fixture for returning middleware."""
        return SetupMiddleware(True)

    @staticmethod
    @patch('django.urls.reverse', return_value='users:login')
    def test_404_urls(reverse, middleware, kwargs):
        """Test how middleware deals with 404 URLs."""
        request = RequestFactory().get('/plinth/non-existing-url')
        response = middleware.process_view(request, **kwargs)
        assert response is None

    @staticmethod
    @patch('django.urls.reverse', return_value='users:login')
    def test_url_not_an_application(reverse, middleware, kwargs):
        """Test that none is returned for URLs that are not applications."""
        request = RequestFactory().get('/plinth/')
        response = middleware.process_view(request, **kwargs)
        assert response is None

    @staticmethod
    @patch('plinth.tests.test_middleware.setup_helper')
    @patch('django.urls.resolve')
    @patch('django.urls.reverse', return_value='users:login')
    def test_module_is_up_to_date(_reverse, resolve, setup_helper_, app,
                                  middleware, kwargs):
        """Test that none is returned when module is up-to-date."""
        resolve.return_value.namespaces = ['mockapp']
        setup_helper_.is_finished = None
        app.get_setup_state = lambda: app_module.App.SetupState.UP_TO_DATE

        request = RequestFactory().get('/plinth/mockapp')
        request.user = AnonymousUser()
        response = middleware.process_view(request, **kwargs)
        assert response is None

    @staticmethod
    @patch('plinth.tests.test_middleware.setup_helper')
    @patch('plinth.views.SetupView')
    @patch('django.urls.resolve')
    @patch('django.urls.reverse', return_value='users:login')
    @pytest.mark.django_db
    def test_module_view(_reverse, resolve, setup_view, setup_helper, app,
                         middleware, kwargs):
        """Test that only registered users can access the setup view."""
        resolve.return_value.namespaces = ['mockapp']
        view = Mock()
        setup_view.as_view.return_value = view
        request = RequestFactory().get('/plinth/mockapp')
        request.session = MagicMock()
        setup_helper.is_finished = None

        # Verify that anonymous users cannot access the setup page
        request.user = AnonymousUser()
        with pytest.raises(PermissionDenied):
            middleware.process_view(request, **kwargs)

        setup_view.as_view.assert_not_called()
        view.assert_not_called()

        # Verify that non-admin logged-in users can't access the setup page
        user = User(username='johndoe')
        user.save()
        request.user = user
        with pytest.raises(PermissionDenied):
            middleware.process_view(request, **kwargs)

        setup_view.as_view.assert_not_called()
        view.assert_not_called()

        # Verify that admin logged-in users can access the setup page
        user = User(username='adminuser')
        user.save()
        group = Group(name='admin')
        group.save()
        user.groups.add(group)
        user.save()
        request.user = user
        middleware.process_view(request, **kwargs)
        setup_view.as_view.assert_called_once_with()
        view.assert_called_once_with(request, app_id='mockapp')

    @staticmethod
    @patch('plinth.operation.manager')
    @patch('django.contrib.messages.error')
    @patch('django.contrib.messages.success')
    @patch('django.urls.resolve')
    @patch('django.urls.reverse', return_value='users:login')
    @pytest.mark.django_db
    def test_install_result_collection(reverse, resolve, messages_success,
                                       messages_error, operation_manager, app,
                                       middleware, kwargs):
        """Test that module installation result is collected properly."""
        resolve.return_value.namespaces = ['mockapp']
        operation_manager.collect_results.return_value = [
            Mock(message='message1', exception=None),
            Mock(message='message2', exception='x-exception')
        ]
        app.get_setup_state = lambda: app_module.App.SetupState.UP_TO_DATE

        # Admin user can collect result
        request = RequestFactory().get('/plinth/mockapp')
        user = User(username='adminuser')
        user.save()
        group = Group(name='admin')
        group.save()
        user.groups.add(group)
        user.save()
        request.user = user
        request.session = MagicMock()
        response = middleware.process_view(request, **kwargs)

        assert response is None
        messages_success.assert_has_calls([call(request, 'message1')])
        messages_error.assert_has_calls([call(request, 'message2')])
        operation_manager.collect_results.assert_has_calls([call('mockapp')])

        # Non-admin user can't collect result
        messages_success.reset_mock()
        messages_error.reset_mock()
        operation_manager.collect_results.reset_mock()
        request = RequestFactory().get('/plinth/mockapp')
        user = User(username='johndoe')
        user.save()
        request.user = user
        request.session = MagicMock()
        response = middleware.process_view(request, **kwargs)

        assert response is None
        assert not messages_success.called
        assert not messages_error.called
        operation_manager.collect_results.assert_not_called()


class TestAdminMiddleware:
    """Test cases for admin middleware."""

    @staticmethod
    @pytest.fixture(name='middleware')
    def fixture_middleware(load_cfg):
        """Fixture for returning middleware."""
        return AdminRequiredMiddleware(True)

    @staticmethod
    @pytest.fixture(name='web_request')
    def fixture_web_request():
        """Fixture for returning kwargs."""
        web_request = RequestFactory().get('/plinth/mockapp')
        web_request.user = Mock()
        return web_request

    @staticmethod
    def test_that_admin_view_is_denied_for_usual_user(web_request, middleware,
                                                      kwargs):
        """Test that normal user is denied for an admin view"""
        web_request.user.groups.filter().exists = Mock(return_value=False)
        web_request.session = MagicMock()
        with pytest.raises(PermissionDenied):
            middleware.process_view(web_request, **kwargs)

    @staticmethod
    def test_group_view_is_denied_for_non_group_user(web_request, middleware,
                                                     kwargs):
        """Test that group view is allowed for an admin user."""
        web_request.user.groups.filter().exists = Mock(return_value=False)
        web_request.session = MagicMock()
        with patch(
                'plinth.middleware.AdminRequiredMiddleware.check_user_group',
                lambda x, y: False):
            with pytest.raises(PermissionDenied):
                middleware.process_view(web_request, **kwargs)

    @staticmethod
    def test_group_view_is_allowed_for_group_user(web_request, middleware,
                                                  kwargs):
        """Test that group view is allowed for an admin user."""
        web_request.user.groups.filter().exists = Mock(return_value=False)
        web_request.session = MagicMock()
        with patch(
                'plinth.middleware.AdminRequiredMiddleware.check_user_group',
                lambda x, y: True):
            response = middleware.process_view(web_request, **kwargs)
            assert response is None

    @staticmethod
    def test_that_admin_view_is_allowed_for_admin_user(web_request, middleware,
                                                       kwargs):
        """Test that admin user is allowed for an admin view"""
        web_request.user.groups.filter().exists = Mock(return_value=True)
        web_request.session = MagicMock()
        response = middleware.process_view(web_request, **kwargs)
        assert response is None

    @staticmethod
    def test_that_public_view_is_allowed_for_normal_user(
            web_request, middleware, kwargs):
        """Test that normal user is allowed for an public view"""
        kwargs = dict(kwargs)
        kwargs['view_func'] = public(HttpResponse)

        response = middleware.process_view(web_request, **kwargs)
        assert response is None


class TestCommonErrorMiddleware:
    """Test cases for common error middleware."""

    @staticmethod
    @pytest.fixture(name='middleware')
    def fixture_middleware(load_cfg):
        """Fixture for returning middleware."""
        return CommonErrorMiddleware(True)

    @staticmethod
    @pytest.fixture(name='web_request')
    def fixture_web_request():
        """Fixture for returning web request."""
        web_request = RequestFactory().get('/plinth/mockapp')
        web_request.user = Mock()
        return web_request

    @staticmethod
    @pytest.fixture(name='operational_error')
    def fixture_operational_error():
        """Fixture for returning an OperationalError."""
        return OperationalError()

    @staticmethod
    @pytest.fixture(name='other_error')
    def fixture_other_error():
        """Fixture for returning a different type of error."""
        return IndexError()

    @staticmethod
    def test_operational_error(middleware, web_request, operational_error):
        response = middleware.process_exception(web_request, operational_error)
        assert response.template_name == 'error.html'
        assert 'message' in response.context_data

    @staticmethod
    def test_other_error(middleware, web_request, other_error):
        response = middleware.process_exception(web_request, other_error)
        assert response is None



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for performance app.
"""

import pytest

from plinth.tests.functional import BaseAppTests

pytestmark = [pytest.mark.system, pytest.mark.performance]


class TestPerformanceApp(BaseAppTests):
    app_name = 'performance'
    has_service = True
    has_web = False



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for Coturn app.
"""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from plinth.modules import coturn
from plinth.modules.coturn.components import TurnConfiguration


def get_domain_choices():
    """Double domain entries for inclusion in the choice field."""
    return ((domain, domain) for domain in coturn.get_available_domains())


def turn_uris_validator(turn_uris):
    """Validate list of STUN/TURN Server URIs."""
    uris = [uri for uri in turn_uris.split('\r\n') if uri]
    if not TurnConfiguration.validate_turn_uris(uris):
        raise ValidationError(_('Invalid list of STUN/TURN Server URIs'))


class CoturnForm(forms.Form):
    """Form to select a TLS domain for Coturn."""

    domain = forms.ChoiceField(
        choices=get_domain_choices,
        label=_('TLS domain'),
        help_text=_(
            'Select a domain to use TLS with. If the list is empty, please '
            'configure at least one domain with certificates.'),
        required=False,
    )



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for using Let's Encrypt.
"""

import logging

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST

from plinth.modules import letsencrypt
from plinth.views import AppView, messages_error

logger = logging.getLogger(__name__)


class LetsEncryptAppView(AppView):
    """Show Let's Encrypt app main page."""

    app_id = 'letsencrypt'
    template_name = 'letsencrypt.html'

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['status'] = letsencrypt.get_status()
        return context


@require_POST
def revoke(request, domain):
    """Revoke a certificate for a given domain."""
    try:
        letsencrypt.certificate_revoke(domain)
        messages.success(
            request,
            _('Certificate successfully revoked for domain {domain}.'
              'This may take a few moments to take effect.').format(
                  domain=domain))
    except Exception as exception:
        messages_error(
            request,
            _('Failed to revoke certificate for domain {domain}').format(
                domain=domain), exception)

    return redirect(reverse_lazy('letsencrypt:index'))


@require_POST
def obtain(request, domain):
    """Obtain and install a certificate for a given domain."""
    try:
        letsencrypt.certificate_obtain(domain)
        messages.success(
            request,
            _('Certificate successfully obtained for domain {domain}').format(
                domain=domain))
    except Exception as exception:
        messages_error(
            request,
            _('Failed to obtain certificate for domain {domain}').format(
                domain=domain), exception)

    return redirect(reverse_lazy('letsencrypt:index'))


@require_POST
def reobtain(request, domain):
    """Re-obtain a certificate for a given domain."""
    try:
        letsencrypt.certificate_reobtain(domain)
        messages.success(
            request,
            _('Certificate successfully obtained for domain {domain}').format(
                domain=domain))
    except Exception as exception:
        messages_error(
            request,
            _('Failed to obtain certificate for domain {domain}').format(
                domain=domain), exception)

    return redirect(reverse_lazy('letsencrypt:index'))


@require_POST
def delete(request, domain):
    """Delete a certificate for a given domain, and cleanup renewal config."""
    try:
        letsencrypt.certificate_delete(domain)
        messages.success(
            request,
            _('Certificate successfully deleted for domain {domain}').format(
                domain=domain))
    except Exception as exception:
        messages_error(
            request,
            _('Failed to delete certificate for domain {domain}').format(
                domain=domain), exception)

    return redirect(reverse_lazy('letsencrypt:index'))
